@@ -9,18 +9,18 @@ account, and subaccounting-excluding and -including balances.
 
 module Hledger.Data.Account
 where
-import Data.List
+import Data.List (find, sortOn)
 import Data.List.Extra (groupSort, groupOn)
-import Data.Maybe
-import Data.Ord
+import Data.Maybe (fromMaybe)
+import Data.Ord (Down(..))
 import qualified Data.Map as M
-import Data.Text (pack,unpack)
+import qualified Data.Text as T
 import Safe (headMay, lookupJustDef)
 import Text.Printf
 
 import Hledger.Data.AccountName
 import Hledger.Data.Amount
-import Hledger.Data.Posting()
+import Hledger.Data.Posting ()
 import Hledger.Data.Types
 import Hledger.Utils
 
@@ -28,11 +28,12 @@ import Hledger.Utils
 -- deriving instance Show Account
 instance Show Account where
     show Account{..} = printf "Account %s (boring:%s, postings:%d, ebalance:%s, ibalance:%s)"
-                       (pack $ regexReplace ":" "_" $ unpack aname)  -- hide : so pretty-show doesn't break line
+                       (T.map colonToUnderscore aname)  -- hide : so pretty-show doesn't break line
                        (if aboring then "y" else "n" :: String)
                        anumpostings
                        (showMixedAmount aebalance)
                        (showMixedAmount aibalance)
+      where colonToUnderscore x = if x == ':' then '_' else x
 
 instance Eq Account where
   (==) a b = aname a == aname b -- quick equality test for speed
@@ -134,11 +135,13 @@ clipAccounts d a = a{asubs=subs}
 
 -- | Remove subaccounts below the specified depth, aggregating their balance at the depth limit
 -- (accounts at the depth limit will have any sub-balances merged into their exclusive balance).
-clipAccountsAndAggregate :: Int -> [Account] -> [Account]
-clipAccountsAndAggregate d as = combined
+-- If the depth is Nothing, return the original accounts
+clipAccountsAndAggregate :: Maybe Int -> [Account] -> [Account]
+clipAccountsAndAggregate Nothing  as = as
+clipAccountsAndAggregate (Just d) as = combined
     where
-      clipped  = [a{aname=clipOrEllipsifyAccountName d $ aname a} | a <- as]
-      combined = [a{aebalance=sum (map aebalance same)}
+      clipped  = [a{aname=clipOrEllipsifyAccountName (Just d) $ aname a} | a <- as]
+      combined = [a{aebalance=sum $ map aebalance same}
                  | same@(a:_) <- groupOn aname clipped]
 {-
 test cases, assuming d=1:
@@ -199,14 +202,11 @@ filterAccounts p a
 -- if balances are normally negative, then the most negative balances
 -- sort first, and vice versa.
 sortAccountTreeByAmount :: NormalSign -> Account -> Account
-sortAccountTreeByAmount normalsign a
-  | null $ asubs a = a
-  | otherwise      = a{asubs=
-                        sortBy (maybeflip $ comparing (normaliseMixedAmountSquashPricesForDisplay . aibalance)) $
-                        map (sortAccountTreeByAmount normalsign) $ asubs a}
+sortAccountTreeByAmount normalsign = mapAccounts $ \a -> a{asubs=sortSubs $ asubs a}
   where
-    maybeflip | normalsign==NormallyNegative = id
-              | otherwise                  = flip
+    sortSubs = case normalsign of
+        NormallyPositive -> sortOn (Down . normaliseMixedAmountSquashPricesForDisplay . aibalance)
+        NormallyNegative -> sortOn (       normaliseMixedAmountSquashPricesForDisplay . aibalance)
 
 -- | Add extra info for this account derived from the Journal's
 -- account directives, if any (comment, tags, declaration order..).
