@@ -16,9 +16,10 @@ module Hledger.Data.AccountName (
   ,accountNameFromComponents
   ,accountNameLevel
   ,accountNameToAccountOnlyRegex
+  ,accountNameToAccountOnlyRegexCI
   ,accountNameToAccountRegex
+  ,accountNameToAccountRegexCI
   ,accountNameTreeFrom
-  ,accountRegexToAccountName
   ,accountSummarisedName
   ,acctsep
   ,acctsepchar
@@ -40,15 +41,14 @@ module Hledger.Data.AccountName (
 )
 where
 
-import Data.List
 import Data.List.Extra (nubSort)
+import qualified Data.List.NonEmpty as NE
 #if !(MIN_VERSION_base(4,11,0))
-import Data.Monoid
+import Data.Semigroup ((<>))
 #endif
 import Data.Text (Text)
 import qualified Data.Text as T
-import Data.Tree
-import Text.Printf
+import Data.Tree (Tree(..))
 
 import Hledger.Data.Types
 import Hledger.Utils
@@ -103,9 +103,11 @@ accountNameDrop n a
       case accountNameDrop n $ T.drop (T.length unbudgetedAccountAndSep) a of
         "" -> unbudgetedAccountName
         a' -> unbudgetedAccountAndSep <> a'
-  | otherwise = accountNameFromComponents $ drop n $ accountNameComponents a
+  | otherwise = accountNameFromComponentsOrElide . drop n $ accountNameComponents a
   where
     unbudgetedAccountAndSep = unbudgetedAccountName <> acctsep
+    accountNameFromComponentsOrElide [] = "..."
+    accountNameFromComponentsOrElide xs = accountNameFromComponents xs
 
 -- | Sorted unique account names implied by these account names,
 -- ie these plus all their parent accounts up to the root.
@@ -115,7 +117,7 @@ expandAccountNames as = nubSort $ concatMap expandAccountName as
 
 -- | "a:b:c" -> ["a","a:b","a:b:c"]
 expandAccountName :: AccountName -> [AccountName]
-expandAccountName = map accountNameFromComponents . tail . inits . accountNameComponents
+expandAccountName = map accountNameFromComponents . NE.tail . NE.inits . accountNameComponents
 
 -- | ["a:b:c","d:e"] -> ["a","d"]
 topAccountNames :: [AccountName] -> [AccountName]
@@ -192,36 +194,45 @@ elideAccountName width s
           | otherwise = done++ss
 
 -- | Keep only the first n components of an account name, where n
--- is a positive integer. If n is 0, returns the empty string.
-clipAccountName :: Int -> AccountName -> AccountName
-clipAccountName n = accountNameFromComponents . take n . accountNameComponents
+-- is a positive integer. If n is Just 0, returns the empty string, if n is
+-- Nothing, return the full name.
+clipAccountName :: Maybe Int -> AccountName -> AccountName
+clipAccountName Nothing  = id
+clipAccountName (Just n) = accountNameFromComponents . take n . accountNameComponents
 
 -- | Keep only the first n components of an account name, where n
--- is a positive integer. If n is 0, returns "...".
-clipOrEllipsifyAccountName :: Int -> AccountName -> AccountName
-clipOrEllipsifyAccountName 0 = const "..."
-clipOrEllipsifyAccountName n = accountNameFromComponents . take n . accountNameComponents
+-- is a positive integer. If n is Just 0, returns "...", if n is Nothing, return
+-- the full name.
+clipOrEllipsifyAccountName :: Maybe Int -> AccountName -> AccountName
+clipOrEllipsifyAccountName (Just 0) = const "..."
+clipOrEllipsifyAccountName n        = clipAccountName n
 
 -- | Escape an AccountName for use within a regular expression.
 -- >>> putStr $ escapeName "First?!#$*?$(*) !@^#*? %)*!@#"
 -- First\?!#\$\*\?\$\(\*\) !@\^#\*\? %\)\*!@#
-escapeName :: AccountName -> Regexp
-escapeName = regexReplaceBy "[[?+|()*\\\\^$]" ("\\" <>)
-           . T.unpack
+escapeName :: AccountName -> String
+escapeName = T.unpack . T.concatMap escapeChar
+  where
+    escapeChar c = if c `elem` escapedChars then T.snoc "\\" c else T.singleton c
+    escapedChars = ['[', '?', '+', '|', '(', ')', '*', '$', '^', '\\']
 
 -- | Convert an account name to a regular expression matching it and its subaccounts.
 accountNameToAccountRegex :: AccountName -> Regexp
-accountNameToAccountRegex "" = ""
-accountNameToAccountRegex a = printf "^%s(:|$)" (escapeName a)
+accountNameToAccountRegex a = toRegex' $ '^' : escapeName a ++ "(:|$)"  -- PARTIAL: Is this safe after escapeName?
+
+-- | Convert an account name to a regular expression matching it and its subaccounts,
+-- case insensitively.
+accountNameToAccountRegexCI :: AccountName -> Regexp
+accountNameToAccountRegexCI a = toRegexCI' $ '^' : escapeName a ++ "(:|$)"  -- PARTIAL: Is this safe after escapeName?
 
 -- | Convert an account name to a regular expression matching it but not its subaccounts.
 accountNameToAccountOnlyRegex :: AccountName -> Regexp
-accountNameToAccountOnlyRegex "" = ""
-accountNameToAccountOnlyRegex a = printf "^%s$"  $ escapeName a -- XXX pack
+accountNameToAccountOnlyRegex a = toRegex' $ '^' : escapeName a ++ "$" -- PARTIAL: Is this safe after escapeName?
 
--- | Convert an exact account-matching regular expression to a plain account name.
-accountRegexToAccountName :: Regexp -> AccountName
-accountRegexToAccountName = T.pack . regexReplace "^\\^(.*?)\\(:\\|\\$\\)$" "\\1" -- XXX pack
+-- | Convert an account name to a regular expression matching it but not its subaccounts,
+-- case insensitively.
+accountNameToAccountOnlyRegexCI :: AccountName -> Regexp
+accountNameToAccountOnlyRegexCI a = toRegexCI' $ '^' : escapeName a ++ "$" -- PARTIAL: Is this safe after escapeName?
 
 -- -- | Does this string look like an exact account-matching regular expression ?
 --isAccountRegex  :: String -> Bool
