@@ -74,7 +74,6 @@ module Hledger.Data.Journal (
   journalCashAccountQuery,
   -- * Misc
   canonicalStyleFrom,
-  matchpats,
   nulljournal,
   journalCheckBalanceAssertions,
   journalNumberAndTieTransactions,
@@ -296,67 +295,55 @@ journalAccountNameTree = accountNameTreeFrom . journalAccountNames
 
 -- queries for standard account types
 
--- | Get a query for accounts of a certain type (Asset, Liability..) in this journal.
--- The query will match all accounts which were declared as that type by account directives,
--- plus all their subaccounts which have not been declared as a different type.
--- If no accounts were declared as this type, the query will instead match accounts
--- with names matched by the provided case-insensitive regular expression.
-journalAccountTypeQuery :: AccountType -> Regexp -> Journal -> Query
-journalAccountTypeQuery atype fallbackregex j =
-  case M.lookup atype (jdeclaredaccounttypes j) of
-    Nothing -> Acct fallbackregex
-    Just as ->
-      -- XXX Query isn't able to match account type since that requires extra info from the journal.
-      -- So we do a hacky search by name instead.
-      And [
-         Or $ map (Acct . accountNameToAccountRegex) as
-        ,Not $ Or $ map (Acct . accountNameToAccountRegex) differentlytypedsubs
-        ]
-      where
-        differentlytypedsubs = concat
-          [subs | (t,bs) <- M.toList (jdeclaredaccounttypes j)
-              , t /= atype
-              , let subs = [b | b <- bs, any (`isAccountNamePrefixOf` b) as]
-          ]
-
 -- | A query for accounts in this journal which have been
--- declared as Asset by account directives, or otherwise for
--- accounts with names matched by the case-insensitive regular expression
--- @^assets?(:|$)@.
+-- declared as Asset (or Cash, a subtype of Asset) by account directives, 
+-- or otherwise for accounts with names matched by the case-insensitive 
+-- regular expression @^assets?(:|$)@.
 journalAssetAccountQuery :: Journal -> Query
-journalAssetAccountQuery = journalAccountTypeQuery Asset "^assets?(:|$)"
+journalAssetAccountQuery = journalAccountTypeQuery [Asset,Cash] (toRegexCI' "^assets?(:|$)")
+
+-- | A query for "Cash" (liquid asset) accounts in this journal, ie accounts
+-- declared as Cash by account directives, or otherwise with names matched by the 
+-- case-insensitive regular expression @^assets?(:|$)@. and not including
+-- the case-insensitive regular expression @(investment|receivable|:A/R|:fixed)@.
+journalCashAccountQuery  :: Journal -> Query
+journalCashAccountQuery j =
+  case M.lookup Cash (jdeclaredaccounttypes j) of
+    Nothing -> And [ journalAssetAccountQuery j, Not . Acct $ toRegexCI' "(investment|receivable|:A/R|:fixed)" ]
+    Just _  -> journalAccountTypeQuery [Cash] notused j
+      where notused = error' "journalCashAccountQuery: this should not have happened!"  -- PARTIAL:
 
 -- | A query for accounts in this journal which have been
 -- declared as Liability by account directives, or otherwise for
 -- accounts with names matched by the case-insensitive regular expression
 -- @^(debts?|liabilit(y|ies))(:|$)@.
 journalLiabilityAccountQuery :: Journal -> Query
-journalLiabilityAccountQuery = journalAccountTypeQuery Liability "^(debts?|liabilit(y|ies))(:|$)"
+journalLiabilityAccountQuery = journalAccountTypeQuery [Liability] (toRegexCI' "^(debts?|liabilit(y|ies))(:|$)")
 
 -- | A query for accounts in this journal which have been
 -- declared as Equity by account directives, or otherwise for
 -- accounts with names matched by the case-insensitive regular expression
 -- @^equity(:|$)@.
 journalEquityAccountQuery :: Journal -> Query
-journalEquityAccountQuery = journalAccountTypeQuery Equity "^equity(:|$)"
+journalEquityAccountQuery = journalAccountTypeQuery [Equity] (toRegexCI' "^equity(:|$)")
 
 -- | A query for accounts in this journal which have been
 -- declared as Revenue by account directives, or otherwise for
 -- accounts with names matched by the case-insensitive regular expression
 -- @^(income|revenue)s?(:|$)@.
 journalRevenueAccountQuery :: Journal -> Query
-journalRevenueAccountQuery = journalAccountTypeQuery Revenue "^(income|revenue)s?(:|$)"
+journalRevenueAccountQuery = journalAccountTypeQuery [Revenue] (toRegexCI' "^(income|revenue)s?(:|$)")
 
 -- | A query for accounts in this journal which have been
 -- declared as Expense by account directives, or otherwise for
 -- accounts with names matched by the case-insensitive regular expression
 -- @^expenses?(:|$)@.
 journalExpenseAccountQuery  :: Journal -> Query
-journalExpenseAccountQuery = journalAccountTypeQuery Expense "^expenses?(:|$)"
+journalExpenseAccountQuery = journalAccountTypeQuery [Expense] (toRegexCI' "^expenses?(:|$)")
 
 -- | A query for Asset, Liability & Equity accounts in this journal.
 -- Cf <http://en.wikipedia.org/wiki/Chart_of_accounts#Balance_Sheet_Accounts>.
-journalBalanceSheetAccountQuery  :: Journal -> Query
+journalBalanceSheetAccountQuery :: Journal -> Query
 journalBalanceSheetAccountQuery j = Or [journalAssetAccountQuery j
                                        ,journalLiabilityAccountQuery j
                                        ,journalEquityAccountQuery j
@@ -369,12 +356,32 @@ journalProfitAndLossAccountQuery j = Or [journalRevenueAccountQuery j
                                         ,journalExpenseAccountQuery j
                                         ]
 
--- | A query for Cash (-equivalent) accounts in this journal (ie,
--- accounts which appear on the cashflow statement.)  This is currently
--- hard-coded to be all the Asset accounts except for those with names
--- containing the case-insensitive regular expression @(receivable|:A/R|:fixed)@.
-journalCashAccountQuery  :: Journal -> Query
-journalCashAccountQuery j = And [journalAssetAccountQuery j, Not $ Acct "(receivable|:A/R|:fixed)"]
+-- | Get a query for accounts of the specified types (Asset, Liability..) in this journal.
+-- The query will match all accounts which were declared as one of
+-- these types by account directives, plus all their subaccounts which
+-- have not been declared as some other type.
+-- Or if no accounts were declared with these types, the query will
+-- instead match accounts with names matched by the provided
+-- case-insensitive regular expression.
+journalAccountTypeQuery :: [AccountType] -> Regexp -> Journal -> Query
+journalAccountTypeQuery atypes fallbackregex Journal{jdeclaredaccounttypes} =
+  let
+    declaredacctsoftype :: [AccountName] =
+      concat $ mapMaybe (`M.lookup` jdeclaredaccounttypes) atypes
+  in case declaredacctsoftype of
+    [] -> Acct fallbackregex
+    as -> And [ Or acctnameRegexes, Not $ Or differentlyTypedRegexes ]
+      where
+        -- XXX Query isn't able to match account type since that requires extra info from the journal.
+        -- So we do a hacky search by name instead.
+        acctnameRegexes = map (Acct . accountNameToAccountRegex) as
+        differentlyTypedRegexes = map (Acct . accountNameToAccountRegex) differentlytypedsubs
+
+        differentlytypedsubs = concat
+          [subs | (t,bs) <- M.toList jdeclaredaccounttypes
+              , not $ t `elem` atypes
+              , let subs = [b | b <- bs, any (`isAccountNamePrefixOf` b) as]
+          ]
 
 -- Various kinds of filtering on journals. We do it differently depending
 -- on the command.
@@ -579,10 +586,15 @@ journalTieTransactions j@Journal{jtxns=ts} = j{jtxns=map txnTieKnot ts}
 journalUntieTransactions :: Transaction -> Transaction
 journalUntieTransactions t@Transaction{tpostings=ps} = t{tpostings=map (\p -> p{ptransaction=Nothing}) ps}
 
--- | Apply any transaction modifier rules in the journal
--- (adding automated postings to transactions, eg).
-journalModifyTransactions :: Journal -> Journal
-journalModifyTransactions j = j{ jtxns = modifyTransactions (jtxnmodifiers j) (jtxns j) }
+-- | Apply any transaction modifier rules in the journal (adding automated
+-- postings to transactions, eg). Or if a modifier rule fails to parse,
+-- return the error message. A reference date is provided to help interpret
+-- relative dates in transaction modifier queries.
+journalModifyTransactions :: Day -> Journal -> Either String Journal
+journalModifyTransactions d j =
+  case modifyTransactions d (jtxnmodifiers j) (jtxns j) of
+    Right ts -> Right j{jtxns=ts}
+    Left err -> Left err
 
 -- | Check any balance assertions in the journal and return an error message
 -- if any of them fail (or if the transaction balancing they require fails).
@@ -700,7 +712,7 @@ journalBalanceTransactions assrt j' =
     runST $ do
       -- We'll update a mutable array of transactions as we balance them,
       -- not strictly necessary but avoids a sort at the end I think.
-      balancedtxns <- newListArray (1, genericLength ts) ts
+      balancedtxns <- newListArray (1, toInteger $ length ts) ts
 
       -- Infer missing posting amounts, check transactions are balanced,
       -- and check balance assertions. This is done in two passes:
@@ -969,7 +981,7 @@ journalInferCommodityStyles :: Journal -> Either String Journal
 journalInferCommodityStyles j =
   case
     commodityStylesFromAmounts $
-    dbg8 "journalInferCommodityStyles using amounts" $
+    dbg7 "journalInferCommodityStyles using amounts" $
     journalStyleInfluencingAmounts j
   of
     Left e   -> Left e
@@ -1221,25 +1233,6 @@ postingFindTag tagname p = find ((tagname==) . fst) $ postingAllTags p
 --     )
 --     ]
 
--- Misc helpers
-
--- | Check if a set of hledger account/description filter patterns matches the
--- given account name or entry description.  Patterns are case-insensitive
--- regular expressions. Prefixed with not:, they become anti-patterns.
-matchpats :: [String] -> String -> Bool
-matchpats pats str =
-    (null positives || any match positives) && (null negatives || not (any match negatives))
-    where
-      (negatives,positives) = partition isnegativepat pats
-      match "" = True
-      match pat = regexMatchesCI (abspat pat) str
-
-negateprefix = "not:"
-
-isnegativepat = (negateprefix `isPrefixOf`)
-
-abspat pat = if isnegativepat pat then drop (length negateprefix) pat else pat
-
 -- debug helpers
 -- traceAmountPrecision a = trace (show $ map (precision . acommodity) $ amounts a) a
 -- tracePostingsCommodities ps = trace (show $ map ((map (precision . acommodity) . amounts) . pamount) ps) ps
@@ -1279,7 +1272,7 @@ Right samplejournal = journalBalanceTransactions False $
            txnTieKnot $ Transaction {
              tindex=0,
              tsourcepos=nullsourcepos,
-             tdate=parsedate "2008/01/01",
+             tdate=fromGregorian 2008 01 01,
              tdate2=Nothing,
              tstatus=Unmarked,
              tcode="",
@@ -1296,7 +1289,7 @@ Right samplejournal = journalBalanceTransactions False $
            txnTieKnot $ Transaction {
              tindex=0,
              tsourcepos=nullsourcepos,
-             tdate=parsedate "2008/06/01",
+             tdate=fromGregorian 2008 06 01,
              tdate2=Nothing,
              tstatus=Unmarked,
              tcode="",
@@ -1313,7 +1306,7 @@ Right samplejournal = journalBalanceTransactions False $
            txnTieKnot $ Transaction {
              tindex=0,
              tsourcepos=nullsourcepos,
-             tdate=parsedate "2008/06/02",
+             tdate=fromGregorian 2008 06 02,
              tdate2=Nothing,
              tstatus=Unmarked,
              tcode="",
@@ -1330,7 +1323,7 @@ Right samplejournal = journalBalanceTransactions False $
            txnTieKnot $ Transaction {
              tindex=0,
              tsourcepos=nullsourcepos,
-             tdate=parsedate "2008/06/03",
+             tdate=fromGregorian 2008 06 03,
              tdate2=Nothing,
              tstatus=Cleared,
              tcode="",
@@ -1347,7 +1340,7 @@ Right samplejournal = journalBalanceTransactions False $
            txnTieKnot $ Transaction {
              tindex=0,
              tsourcepos=nullsourcepos,
-             tdate=parsedate "2008/10/01",
+             tdate=fromGregorian 2008 10 01,
              tdate2=Nothing,
              tstatus=Unmarked,
              tcode="",
@@ -1363,7 +1356,7 @@ Right samplejournal = journalBalanceTransactions False $
            txnTieKnot $ Transaction {
              tindex=0,
              tsourcepos=nullsourcepos,
-             tdate=parsedate "2008/12/31",
+             tdate=fromGregorian 2008 12 31,
              tdate2=Nothing,
              tstatus=Unmarked,
              tcode="",
@@ -1382,11 +1375,11 @@ tests_Journal = tests "Journal" [
 
    test "journalDateSpan" $
     journalDateSpan True nulljournal{
-      jtxns = [nulltransaction{tdate = parsedate "2014/02/01"
-                              ,tpostings = [posting{pdate=Just (parsedate "2014/01/10")}]
+      jtxns = [nulltransaction{tdate = fromGregorian 2014 02 01
+                              ,tpostings = [posting{pdate=Just (fromGregorian 2014 01 10)}]
                               }
-              ,nulltransaction{tdate = parsedate "2014/09/01"
-                              ,tpostings = [posting{pdate2=Just (parsedate "2014/10/10")}]
+              ,nulltransaction{tdate = fromGregorian 2014 09 01
+                              ,tpostings = [posting{pdate2=Just (fromGregorian 2014 10 10)}]
                               }
               ]
       }
@@ -1399,11 +1392,18 @@ tests_Journal = tests "Journal" [
       journalAccountNamesMatching q = filter (q `matchesAccount`) . journalAccountNames
       namesfrom qfunc = journalAccountNamesMatching (qfunc j) j
     in [
-       test "assets"      $ assertEqual "" (namesfrom journalAssetAccountQuery)     ["assets","assets:bank","assets:bank:checking","assets:bank:saving","assets:cash"]
-      ,test "liabilities" $ assertEqual "" (namesfrom journalLiabilityAccountQuery) ["liabilities","liabilities:debts"]
-      ,test "equity"      $ assertEqual "" (namesfrom journalEquityAccountQuery)    []
-      ,test "income"      $ assertEqual "" (namesfrom journalRevenueAccountQuery)    ["income","income:gifts","income:salary"]
-      ,test "expenses"    $ assertEqual "" (namesfrom journalExpenseAccountQuery)   ["expenses","expenses:food","expenses:supplies"]
+       test "assets"      $ assertEqual "" ["assets","assets:bank","assets:bank:checking","assets:bank:saving","assets:cash"]
+         (namesfrom journalAssetAccountQuery)
+      ,test "cash"        $ assertEqual "" ["assets","assets:bank","assets:bank:checking","assets:bank:saving","assets:cash"]
+        (namesfrom journalCashAccountQuery)
+      ,test "liabilities" $ assertEqual "" ["liabilities","liabilities:debts"]
+        (namesfrom journalLiabilityAccountQuery)
+      ,test "equity"      $ assertEqual "" []
+        (namesfrom journalEquityAccountQuery)
+      ,test "income"      $ assertEqual "" ["income","income:gifts","income:salary"]
+        (namesfrom journalRevenueAccountQuery)
+      ,test "expenses"    $ assertEqual "" ["expenses","expenses:food","expenses:supplies"]
+        (namesfrom journalExpenseAccountQuery)
     ]
 
   ,tests "journalBalanceTransactions" [
@@ -1413,7 +1413,7 @@ tests_Journal = tests "Journal" [
             --2019/01/01
             --  (a)            = 1
             nulljournal{ jtxns = [
-              transaction "2019/01/01" [ vpost' "a" missingamt (balassert (num 1)) ]
+              transaction (fromGregorian 2019 01 01) [ vpost' "a" missingamt (balassert (num 1)) ]
             ]}
       assertRight ej
       let Right j = ej
@@ -1426,8 +1426,8 @@ tests_Journal = tests "Journal" [
             --2019/01/01
             --  (a)          1 = 2
             nulljournal{ jtxns = [
-               transaction "2019/01/01" [ vpost' "a" missingamt (balassert (num 1)) ]
-              ,transaction "2019/01/01" [ vpost' "a" (num 1)    (balassert (num 2)) ]
+               transaction (fromGregorian 2019 01 01) [ vpost' "a" missingamt (balassert (num 1)) ]
+              ,transaction (fromGregorian 2019 01 01) [ vpost' "a" (num 1)    (balassert (num 2)) ]
             ]}
 
     ,test "same-day-2" $ do
@@ -1440,12 +1440,12 @@ tests_Journal = tests "Journal" [
             --2019/01/01
             --    a                    0 = 1
             nulljournal{ jtxns = [
-               transaction "2019/01/01" [ vpost' "a" (num 2)    (balassert (num 2)) ]
-              ,transaction "2019/01/01" [
+               transaction (fromGregorian 2019 01 01) [ vpost' "a" (num 2)    (balassert (num 2)) ]
+              ,transaction (fromGregorian 2019 01 01) [
                  post' "b" (num 1)     Nothing
                 ,post' "a"  missingamt Nothing
               ]
-              ,transaction "2019/01/01" [ post' "a" (num 0)     (balassert (num 1)) ]
+              ,transaction (fromGregorian 2019 01 01) [ post' "a" (num 0)     (balassert (num 1)) ]
             ]}
 
     ,test "out-of-order" $ do
@@ -1455,8 +1455,8 @@ tests_Journal = tests "Journal" [
             --2019/1/1
             --  (a)    1 = 1
             nulljournal{ jtxns = [
-               transaction "2019/01/02" [ vpost' "a" (num 1)    (balassert (num 2)) ]
-              ,transaction "2019/01/01" [ vpost' "a" (num 1)    (balassert (num 1)) ]
+               transaction (fromGregorian 2019 01 02) [ vpost' "a" (num 1)    (balassert (num 2)) ]
+              ,transaction (fromGregorian 2019 01 01) [ vpost' "a" (num 1)    (balassert (num 1)) ]
             ]}
 
     ]
@@ -1472,26 +1472,26 @@ tests_Journal = tests "Journal" [
       --
       test "1091a" $ do
         commodityStylesFromAmounts [
-           nullamt{aquantity=1000, astyle=AmountStyle L False 3 (Just ',') Nothing}
-          ,nullamt{aquantity=1000, astyle=AmountStyle L False 2 (Just '.') (Just (DigitGroups ',' [3]))}
+           nullamt{aquantity=1000, astyle=AmountStyle L False (Precision 3) (Just ',') Nothing}
+          ,nullamt{aquantity=1000, astyle=AmountStyle L False (Precision 2) (Just '.') (Just (DigitGroups ',' [3]))}
           ]
          @?=
           -- The commodity style should have period as decimal mark
           -- and comma as digit group mark.
           Right (M.fromList [
-            ("", AmountStyle L False 3 (Just '.') (Just (DigitGroups ',' [3])))
+            ("", AmountStyle L False (Precision 3) (Just '.') (Just (DigitGroups ',' [3])))
           ])
         -- same journal, entries in reverse order
       ,test "1091b" $ do
         commodityStylesFromAmounts [
-           nullamt{aquantity=1000, astyle=AmountStyle L False 2 (Just '.') (Just (DigitGroups ',' [3]))}
-          ,nullamt{aquantity=1000, astyle=AmountStyle L False 3 (Just ',') Nothing}
+           nullamt{aquantity=1000, astyle=AmountStyle L False (Precision 2) (Just '.') (Just (DigitGroups ',' [3]))}
+          ,nullamt{aquantity=1000, astyle=AmountStyle L False (Precision 3) (Just ',') Nothing}
           ]
          @?=
           -- The commodity style should have period as decimal mark
           -- and comma as digit group mark.
           Right (M.fromList [
-            ("", AmountStyle L False 3 (Just '.') (Just (DigitGroups ',' [3])))
+            ("", AmountStyle L False (Precision 3) (Just '.') (Just (DigitGroups ',' [3])))
           ])
 
      ]
