@@ -9,7 +9,7 @@ looking up historical market prices (exchange rates) between commodities.
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE RecordWildCards #-}
 {-# LANGUAGE ScopedTypeVariables #-}
-{-# LANGUAGE DeriveDataTypeable, DeriveGeneric #-}
+{-# LANGUAGE DeriveGeneric #-}
 
 module Hledger.Data.Valuation (
    ValuationType(..)
@@ -28,8 +28,6 @@ module Hledger.Data.Valuation (
 where
 
 import Control.Applicative ((<|>))
-import Control.DeepSeq (NFData)
-import Data.Data
 import Data.Decimal (roundTo)
 import Data.Function ((&), on)
 import Data.Graph.Inductive  (Gr, Node, NodeMap, mkMapGraph, mkNode, lab, out, sp)
@@ -38,7 +36,7 @@ import Data.List.Extra (nubSortBy)
 import qualified Data.Map as M
 import Data.Maybe
 import qualified Data.Text as T
-import Data.Time.Calendar (Day)
+import Data.Time.Calendar (Day, fromGregorian)
 import Data.MemoUgly (memo)
 import GHC.Generics (Generic)
 import Safe (headMay)
@@ -46,7 +44,6 @@ import Safe (headMay)
 import Hledger.Utils
 import Hledger.Data.Types
 import Hledger.Data.Amount
-import Hledger.Data.Dates (parsedate)
 
 
 ------------------------------------------------------------------------------
@@ -61,7 +58,7 @@ data ValuationType =
   | AtNow      (Maybe CommoditySymbol)  -- ^ convert to default or given valuation commodity, using current market prices
   | AtDate Day (Maybe CommoditySymbol)  -- ^ convert to default or given valuation commodity, using market prices on some date
   | AtDefault  (Maybe CommoditySymbol)  -- ^ works like AtNow in single period reports, like AtEnd in multiperiod reports
-  deriving (Show,Data,Eq) -- Typeable
+  deriving (Show,Eq)
 
 -- | A snapshot of the known exchange rates between commodity pairs at a given date,
 -- as a graph allowing fast lookup and path finding, along with some helper data.
@@ -88,8 +85,6 @@ data PriceGraph = PriceGraph {
   }
   deriving (Show,Generic)
 
-instance NFData PriceGraph
-
 -- | A price oracle is a magic memoising function that efficiently
 -- looks up market prices (exchange rates) from one commodity to
 -- another (or if unspecified, to a default valuation commodity) on a
@@ -167,7 +162,7 @@ amountApplyValuation priceoracle styles periodlast mreportlast today ismultiperi
   case v of
     AtCost    Nothing            -> styleAmount styles $ amountCost a
     AtCost    mc                 -> amountValueAtDate priceoracle styles mc periodlast $ styleAmount styles $ amountCost a
-    AtThen    _mc                -> error' unsupportedValueThenError  -- TODO
+    AtThen    _mc                -> error' unsupportedValueThenError  -- PARTIAL:
                                  -- amountValueAtDate priceoracle styles mc periodlast a  -- posting date unknown, handle like AtEnd
     AtEnd     mc                 -> amountValueAtDate priceoracle styles mc periodlast a
     AtNow     mc                 -> amountValueAtDate priceoracle styles mc today a
@@ -258,6 +253,8 @@ priceLookup makepricegraph d from mto =
               Nothing    -> Nothing
               Just nodes ->
                 dbg ("market price for "++intercalate " -> " (map T.unpack comms)) $
+                  -- TODO: it would be nice to include price date as part of the label
+                  -- in PriceGraph, so we could show the dates of market prices here
                 Just $ product $ pathEdgeLabels g nodes  -- convert to a single exchange rate
                 where comms = catMaybes $ map (lab g) nodes
 
@@ -266,21 +263,20 @@ priceLookup makepricegraph d from mto =
 
 tests_priceLookup =
   let
-    d = parsedate
-    p date from q to = MarketPrice{mpdate=d date, mpfrom=from, mpto=to, mprate=q}
+    p y m d from q to = MarketPrice{mpdate=fromGregorian y m d, mpfrom=from, mpto=to, mprate=q}
     ps1 = [
-       p "2000/01/01" "A" 10 "B"
-      ,p "2000/01/01" "B" 10 "C"
-      ,p "2000/01/01" "C" 10 "D"
-      ,p "2000/01/01" "E"  2 "D"
-      ,p "2001/01/01" "A" 11 "B"
+       p 2000 01 01 "A" 10 "B"
+      ,p 2000 01 01 "B" 10 "C"
+      ,p 2000 01 01 "C" 10 "D"
+      ,p 2000 01 01 "E"  2 "D"
+      ,p 2001 01 01 "A" 11 "B"
       ]
     makepricegraph = makePriceGraph ps1 []
   in test "priceLookup" $ do
-    priceLookup makepricegraph (d "1999/01/01") "A" Nothing    @?= Nothing
-    priceLookup makepricegraph (d "2000/01/01") "A" Nothing    @?= Just ("B",10)
-    priceLookup makepricegraph (d "2000/01/01") "B" (Just "A") @?= Just ("A",0.1)
-    priceLookup makepricegraph (d "2000/01/01") "A" (Just "E") @?= Just ("E",500)
+    priceLookup makepricegraph (fromGregorian 1999 01 01) "A" Nothing    @?= Nothing
+    priceLookup makepricegraph (fromGregorian 2000 01 01) "A" Nothing    @?= Just ("B",10)
+    priceLookup makepricegraph (fromGregorian 2000 01 01) "B" (Just "A") @?= Just ("A",0.1)
+    priceLookup makepricegraph (fromGregorian 2000 01 01) "A" (Just "E") @?= Just ("E",500)
 
 -- | Build the graph of commodity conversion prices for a given day.
 -- Converts a list of declared market prices in parse order, and a
@@ -400,7 +396,7 @@ node m = fst . fst . mkNode m
 -- lowest-sorting label is used.
 pathEdgeLabels :: (Show b, Ord b) => Gr a b -> [Node] -> [b]
 pathEdgeLabels g = map frommaybe . map (nodesEdgeLabel g) . pathEdges
-  where frommaybe = fromMaybe (error' "pathEdgeLabels: expected no Nothings here")
+  where frommaybe = fromMaybe (error' "pathEdgeLabels: expected no Nothings here")  -- PARTIAL:
 
 -- | Convert a path to node pairs representing the path's edges.
 pathEdges :: [Node] -> [(Node,Node)]
