@@ -12,20 +12,19 @@ Some of these might belong in Hledger.Read.JournalReader or Hledger.Read.
 -}
 
 --- ** language
-{-# LANGUAGE BangPatterns #-}
-{-# LANGUAGE CPP #-}
-{-# LANGUAGE DeriveDataTypeable #-}
-{-# LANGUAGE FlexibleContexts #-}
-{-# LANGUAGE LambdaCase #-}
-{-# LANGUAGE NamedFieldPuns #-}
-{-# LANGUAGE NoMonoLocalBinds #-}
-{-# LANGUAGE OverloadedStrings #-}
-{-# LANGUAGE PackageImports #-}
-{-# LANGUAGE Rank2Types #-}
-{-# LANGUAGE RecordWildCards #-}
+{-# LANGUAGE BangPatterns        #-}
+{-# LANGUAGE CPP                 #-}
+{-# LANGUAGE FlexibleContexts    #-}
+{-# LANGUAGE LambdaCase          #-}
+{-# LANGUAGE NamedFieldPuns      #-}
+{-# LANGUAGE NoMonoLocalBinds    #-}
+{-# LANGUAGE OverloadedStrings   #-}
+{-# LANGUAGE PackageImports      #-}
+{-# LANGUAGE Rank2Types          #-}
+{-# LANGUAGE RecordWildCards     #-}
 {-# LANGUAGE ScopedTypeVariables #-}
-{-# LANGUAGE TupleSections #-}
-{-# LANGUAGE TypeFamilies #-}
+{-# LANGUAGE TupleSections       #-}
+{-# LANGUAGE TypeFamilies        #-}
 
 --- ** exports
 module Hledger.Read.Common (
@@ -105,6 +104,9 @@ module Hledger.Read.Common (
   singlespacedtextsatisfyingp,
   singlespacep,
 
+  skipNonNewlineSpaces,
+  skipNonNewlineSpaces1,
+
   -- * tests
   tests_Common,
 )
@@ -113,31 +115,33 @@ where
 --- ** imports
 import Prelude ()
 import "base-compat-batteries" Prelude.Compat hiding (fail, readFile)
+import Control.Applicative.Permutations (runPermutation, toPermutationWithDefault)
 import qualified "base-compat-batteries" Control.Monad.Fail.Compat as Fail (fail)
 import Control.Monad.Except (ExceptT(..), runExceptT, throwError)
 import Control.Monad.State.Strict hiding (fail)
 import Data.Bifunctor (bimap, second)
-import Data.Char
-import Data.Data
+import Data.Char (digitToInt, isDigit, isSpace)
 import Data.Decimal (DecimalRaw (Decimal), Decimal)
-import Data.Default
+import Data.Default (Default(..))
 import Data.Function ((&))
-import Data.Functor.Identity
+import Data.Functor.Identity (Identity)
 import "base-compat-batteries" Data.List.Compat
 import Data.List.NonEmpty (NonEmpty(..))
-import Data.Maybe
+import Data.Maybe (catMaybes, fromMaybe, isJust, listToMaybe)
 import qualified Data.Map as M
 import qualified Data.Semigroup as Sem
 import Data.Text (Text)
 import qualified Data.Text as T
-import Data.Time.Calendar
-import Data.Time.LocalTime
+import Data.Time.Calendar (Day, fromGregorianValid, toGregorian)
+import Data.Time.LocalTime (LocalTime(..), TimeOfDay(..))
+import Data.Word (Word8)
 import System.Time (getClockTime)
 import Text.Megaparsec
-import Text.Megaparsec.Char
+import Text.Megaparsec.Char (char, char', digitChar, newline, string)
 import Text.Megaparsec.Char.Lexer (decimal)
 import Text.Megaparsec.Custom
-import Control.Applicative.Permutations
+  (FinalParseError, attachSource, customErrorBundlePretty,
+  finalErrorBundlePretty, parseErrorAt, parseErrorAtRegion)
 
 import Hledger.Data
 import Hledger.Utils
@@ -190,7 +194,7 @@ data InputOpts = InputOpts {
     ,new_save_          :: Bool                 -- ^ save latest new transactions state for next time
     ,pivot_             :: String               -- ^ use the given field's value as the account name
     ,auto_              :: Bool                 -- ^ generate automatic postings when journal is parsed
- } deriving (Show, Data) --, Typeable)
+ } deriving (Show)
 
 instance Default InputOpts where def = definputopts
 
@@ -237,14 +241,13 @@ runErroringJournalParser p t =
 rejp = runErroringJournalParser
 
 genericSourcePos :: SourcePos -> GenericSourcePos
-genericSourcePos p = GenericSourcePos (sourceName p) (fromIntegral . unPos $ sourceLine p) (fromIntegral . unPos $ sourceColumn p)
+genericSourcePos p = GenericSourcePos (sourceName p) (unPos $ sourceLine p) (unPos $ sourceColumn p)
 
 -- | Construct a generic start & end line parse position from start and end megaparsec SourcePos's.
 journalSourcePos :: SourcePos -> SourcePos -> GenericSourcePos
-journalSourcePos p p' = JournalSourcePos (sourceName p) (fromIntegral . unPos $ sourceLine p, fromIntegral $ line')
-    where line'
-            | (unPos $ sourceColumn p') == 1 = unPos (sourceLine p') - 1
-            | otherwise = unPos $ sourceLine p' -- might be at end of file withat last new-line
+journalSourcePos p p' = JournalSourcePos (sourceName p) (unPos $ sourceLine p, line')
+    where line' | (unPos $ sourceColumn p') == 1 = unPos (sourceLine p') - 1
+                | otherwise = unPos $ sourceLine p' -- might be at end of file withat last new-line
 
 -- | Given a parser to ParsedJournal, input options, file path and
 -- content: run the parser on the content, and finalise the result to
@@ -295,6 +298,7 @@ parseAndFinaliseJournal' parser iopts f txt = do
 journalFinalise :: InputOpts -> FilePath -> Text -> Journal -> ExceptT String IO Journal
 journalFinalise iopts f txt pj = do
   t <- liftIO getClockTime
+  d <- liftIO getCurrentDay
   -- Infer and apply canonical styles for each commodity (or fail).
   -- This affects transaction balancing/assertions/assignments, so needs to be done early.
   -- (TODO: since #903's refactoring for hledger 1.12,
@@ -319,11 +323,13 @@ journalFinalise iopts f txt pj = do
            -- then add the auto postings
            -- (Note adding auto postings after balancing means #893b fails;
            -- adding them before balancing probably means #893a, #928, #938 fail.)
-           let j'' = journalModifyTransactions j'
-           -- then apply commodity styles once more, to style the auto posting amounts. (XXX inefficient ?)
-           j''' <- journalApplyCommodityStyles j''
-           -- then check balance assertions.
-           journalBalanceTransactions (not $ ignore_assertions_ iopts) j'''
+           case journalModifyTransactions d j' of
+             Left e -> throwError e
+             Right j'' -> do
+               -- then apply commodity styles once more, to style the auto posting amounts. (XXX inefficient ?)
+               j''' <- journalApplyCommodityStyles j''
+               -- then check balance assertions.
+               journalBalanceTransactions (not $ ignore_assertions_ iopts) j'''
         )
      & fmap journalInferMarketPricesFromTransactions  -- infer market prices from commodity-exchanging transactions
 
@@ -412,15 +418,15 @@ match' p = do
 statusp :: TextParser m Status
 statusp =
   choice'
-    [ skipMany spacenonewline >> char '*' >> return Cleared
-    , skipMany spacenonewline >> char '!' >> return Pending
+    [ skipNonNewlineSpaces >> char '*' >> return Cleared
+    , skipNonNewlineSpaces >> char '!' >> return Pending
     , return Unmarked
     ]
 
 codep :: TextParser m Text
 codep = option "" $ do
   try $ do
-    skipSome spacenonewline
+    skipNonNewlineSpaces1
     char '('
   code <- takeWhileP Nothing $ \c -> c /= ')' && c /= '\n'
   char ')' <?> "closing bracket ')' for transaction code"
@@ -443,45 +449,44 @@ datep = do
 
 datep' :: Maybe Year -> TextParser m Day
 datep' mYear = do
-  startOffset <- getOffset
-  d1 <- decimal <?> "year or month"
-  sep <- satisfy isDateSepChar <?> "date separator"
-  d2 <- decimal <?> "month or day"
-  fullDate startOffset d1 sep d2 <|> partialDate startOffset mYear d1 sep d2
-  <?> "full or partial date"
-
+    startOffset <- getOffset
+    d1 <- yearorintp <?> "year or month"
+    sep <- datesepchar <?> "date separator"
+    d2 <- decimal <?> "month or day"
+    case d1 of
+         Left y  -> fullDate startOffset y sep d2
+         Right m -> partialDate startOffset mYear m sep d2
+    <?> "full or partial date"
   where
+    fullDate :: Int -> Year -> Char -> Month -> TextParser m Day
+    fullDate startOffset year sep1 month = do
+      sep2 <- satisfy isDateSepChar <?> "date separator"
+      day <- decimal <?> "day"
+      endOffset <- getOffset
+      let dateStr = show year ++ [sep1] ++ show month ++ [sep2] ++ show day
 
-  fullDate :: Int -> Integer -> Char -> Int -> TextParser m Day
-  fullDate startOffset year sep1 month = do
-    sep2 <- satisfy isDateSepChar <?> "date separator"
-    day <- decimal <?> "day"
-    endOffset <- getOffset
-    let dateStr = show year ++ [sep1] ++ show month ++ [sep2] ++ show day
+      when (sep1 /= sep2) $ customFailure $ parseErrorAtRegion startOffset endOffset $
+        "invalid date (mixing date separators is not allowed): " ++ dateStr
 
-    when (sep1 /= sep2) $ customFailure $ parseErrorAtRegion startOffset endOffset $
-      "invalid date (mixing date separators is not allowed): " ++ dateStr
+      case fromGregorianValid year month day of
+        Nothing -> customFailure $ parseErrorAtRegion startOffset endOffset $
+                     "well-formed but invalid date: " ++ dateStr
+        Just date -> pure $! date
 
-    case fromGregorianValid year month day of
-      Nothing -> customFailure $ parseErrorAtRegion startOffset endOffset $
-                   "well-formed but invalid date: " ++ dateStr
-      Just date -> pure $! date
+    partialDate :: Int -> Maybe Year -> Month -> Char -> MonthDay -> TextParser m Day
+    partialDate startOffset mYear month sep day = do
+      endOffset <- getOffset
+      case mYear of
+        Just year ->
+          case fromGregorianValid year month day of
+            Nothing -> customFailure $ parseErrorAtRegion startOffset endOffset $
+                        "well-formed but invalid date: " ++ dateStr
+            Just date -> pure $! date
+          where dateStr = show year ++ [sep] ++ show month ++ [sep] ++ show day
 
-  partialDate
-    :: Int -> Maybe Year -> Integer -> Char -> Int -> TextParser m Day
-  partialDate startOffset mYear month sep day = do
-    endOffset <- getOffset
-    case mYear of
-      Just year ->
-        case fromGregorianValid year (fromIntegral month) day of
-          Nothing -> customFailure $ parseErrorAtRegion startOffset endOffset $
-                      "well-formed but invalid date: " ++ dateStr
-          Just date -> pure $! date
-        where dateStr = show year ++ [sep] ++ show month ++ [sep] ++ show day
-
-      Nothing -> customFailure $ parseErrorAtRegion startOffset endOffset $
-        "partial date "++dateStr++" found, but the current year is unknown"
-        where dateStr = show month ++ [sep] ++ show day
+        Nothing -> customFailure $ parseErrorAtRegion startOffset endOffset $
+          "partial date "++dateStr++" found, but the current year is unknown"
+          where dateStr = show month ++ [sep] ++ show day
 
 {-# INLINABLE datep' #-}
 
@@ -499,7 +504,7 @@ datetimep = do
 datetimep' :: Maybe Year -> TextParser m LocalTime
 datetimep' mYear = do
   day <- datep' mYear
-  skipSome spacenonewline
+  skipNonNewlineSpaces1
   time <- timeOfDay
   optional timeZone -- ignoring time zones
   pure $ LocalTime day time
@@ -548,21 +553,36 @@ secondarydatep :: Day -> TextParser m Day
 secondarydatep primaryDate = char '=' *> datep' (Just primaryYear)
   where primaryYear = first3 $ toGregorian primaryDate
 
+-- | Parse a year number or an Int. Years must contain at least four
+-- digits.
+yearorintp :: TextParser m (Either Year Int)
+yearorintp = do
+    yearOrMonth <- takeWhile1P (Just "digit") isDigit
+    let n = readDecimal yearOrMonth
+    return $ if T.length yearOrMonth >= 4 then Left n else Right (fromInteger n)
+
 --- *** account names
 
 -- | Parse an account name (plus one following space if present),
 -- then apply any parent account prefix and/or account aliases currently in effect,
 -- in that order. (Ie first add the parent account prefix, then rewrite with aliases).
+-- This calls error if any account alias with an invalid regular expression exists.
 modifiedaccountnamep :: JournalParser m AccountName
 modifiedaccountnamep = do
-  parent <- getParentAccount
+  parent  <- getParentAccount
   aliases <- getAccountAliases
-  a <- lift accountnamep
-  return $!
-    accountNameApplyAliases aliases $
-     -- XXX accountNameApplyAliasesMemo ? doesn't seem to make a difference (retest that function)
-    joinAccountNames parent
-    a
+  -- off1    <- getOffset
+  a       <- lift accountnamep
+  -- off2    <- getOffset
+  -- XXX or accountNameApplyAliasesMemo ? doesn't seem to make a difference (retest that function)
+  case accountNameApplyAliases aliases $ joinAccountNames parent a of
+    Right a' -> return $! a'
+    -- should not happen, regexaliasp will have displayed a better error already:
+    -- (XXX why does customFailure cause error to be displayed there, but not here ?)
+    -- Left e  -> customFailure $! parseErrorAtRegion off1 off2 err
+    Left e   -> error' err  -- PARTIAL:
+      where
+        err = "problem in account alias applied to "++T.unpack a++": "++e
 
 -- | Parse an account name, plus one following space if present.
 -- Account names have one or more parts separated by the account separator character,
@@ -595,7 +615,7 @@ singlespacedtextsatisfyingp pred = do
 
 -- | Parse one non-newline whitespace character that is not followed by another one.
 singlespacep :: TextParser m ()
-singlespacep = void spacenonewline *> notFollowedBy spacenonewline
+singlespacep = spacenonewline *> notFollowedBy spacenonewline
 
 --- *** amounts
 
@@ -605,7 +625,7 @@ singlespacep = void spacenonewline *> notFollowedBy spacenonewline
 spaceandamountormissingp :: JournalParser m MixedAmount
 spaceandamountormissingp =
   option missingmixedamt $ try $ do
-    lift $ skipSome spacenonewline
+    lift $ skipNonNewlineSpaces1
     Mixed . (:[]) <$> amountp
 
 -- | Parse a single-commodity amount, with optional symbol on the left
@@ -614,7 +634,7 @@ spaceandamountormissingp =
 -- lot date. A lot price and lot date will be ignored.
 amountp :: JournalParser m Amount
 amountp = label "amount" $ do
-  let spaces = lift $ skipMany spacenonewline
+  let spaces = lift $ skipNonNewlineSpaces
   amount <- amountwithoutpricep <* spaces
   (mprice, _elotprice, _elotdate) <- runPermutation $
     (,,) <$> toPermutationWithDefault Nothing (Just <$> priceamountp <* spaces)
@@ -625,7 +645,7 @@ amountp = label "amount" $ do
 -- XXX Just like amountp but don't allow lot prices. Needed for balanceassertionp.
 amountpnolotprices :: JournalParser m Amount
 amountpnolotprices = label "amount" $ do
-  let spaces = lift $ skipMany spacenonewline
+  let spaces = lift $ skipNonNewlineSpaces
   amount <- amountwithoutpricep
   spaces
   mprice <- optional $ priceamountp <* spaces
@@ -642,7 +662,7 @@ amountwithoutpricep = do
   leftsymbolamountp mult sign = label "amount" $ do
     c <- lift commoditysymbolp
     suggestedStyle <- getAmountStyle c
-    commodityspaced <- lift $ skipMany' spacenonewline
+    commodityspaced <- lift skipNonNewlineSpaces'
     sign2 <- lift $ signp
     offBeforeNum <- getOffset
     ambiguousRawNum <- lift rawnumberp
@@ -660,7 +680,7 @@ amountwithoutpricep = do
     mExponent <- lift $ optional $ try exponentp
     offAfterNum <- getOffset
     let numRegion = (offBeforeNum, offAfterNum)
-    mSpaceAndCommodity <- lift $ optional $ try $ (,) <$> skipMany' spacenonewline <*> commoditysymbolp
+    mSpaceAndCommodity <- lift $ optional $ try $ (,) <$> skipNonNewlineSpaces' <*> commoditysymbolp
     case mSpaceAndCommodity of
       -- right symbol amount
       Just (commodityspaced, c) -> do
@@ -686,21 +706,21 @@ amountwithoutpricep = do
     :: (Int, Int) -- offsets
     -> Maybe AmountStyle
     -> Either AmbiguousNumber RawNumber
-    -> Maybe Int
-    -> TextParser m (Quantity, Int, Maybe Char, Maybe DigitGroupStyle)
+    -> Maybe Integer
+    -> TextParser m (Quantity, AmountPrecision, Maybe Char, Maybe DigitGroupStyle)
   interpretNumber posRegion suggestedStyle ambiguousNum mExp =
     let rawNum = either (disambiguateNumber suggestedStyle) id ambiguousNum
     in  case fromRawNumber rawNum mExp of
           Left errMsg -> customFailure $
                            uncurry parseErrorAtRegion posRegion errMsg
-          Right res -> pure res
+          Right (q,p,d,g) -> pure (q, Precision p, d, g)
 
 -- | Parse an amount from a string, or get an error.
 amountp' :: String -> Amount
 amountp' s =
   case runParser (evalStateT (amountp <* eof) nulljournal) "" (T.pack s) of
     Right amt -> amt
-    Left err  -> error' $ show err -- XXX should throwError
+    Left err  -> error' $ show err  -- PARTIAL: XXX should throwError
 
 -- | Parse a mixed amount from a string, or get an error.
 mamountp' :: String -> MixedAmount
@@ -709,23 +729,11 @@ mamountp' = Mixed . (:[]) . amountp'
 -- | Parse a minus or plus sign followed by zero or more spaces,
 -- or nothing, returning a function that negates or does nothing.
 signp :: Num a => TextParser m (a -> a)
-signp = ((char '-' *> pure negate <|> char '+' *> pure id) <* many spacenonewline) <|> pure id
+signp = ((char '-' *> pure negate <|> char '+' *> pure id) <* skipNonNewlineSpaces) <|> pure id
 
 multiplierp :: TextParser m Bool
 multiplierp = option False $ char '*' *> pure True
 
--- | This is like skipMany but it returns True if at least one element
--- was skipped. This is helpful if you’re just using many to check if
--- the resulting list is empty or not.
-skipMany' :: MonadPlus m => m a -> m Bool
-skipMany' p = go False
-  where
-    go !isNull = do
-      more <- option False (True <$ p)
-      if more
-        then go True
-        else pure isNull
-
 commoditysymbolp :: TextParser m CommoditySymbol
 commoditysymbolp =
   quotedcommoditysymbolp <|> simplecommoditysymbolp <?> "commodity symbol"
@@ -746,7 +754,7 @@ priceamountp = label "transaction price" $ do
   priceConstructor <- char '@' *> pure TotalPrice <|> pure UnitPrice
   when parenthesised $ void $ char ')'
 
-  lift (skipMany spacenonewline)
+  lift skipNonNewlineSpaces
   priceAmount <- amountwithoutpricep -- <?> "unpriced amount (specifying a price)"
 
   pure $ priceConstructor priceAmount
@@ -757,7 +765,7 @@ balanceassertionp = do
   char '='
   istotal <- fmap isJust $ optional $ try $ char '='
   isinclusive <- fmap isJust $ optional $ try $ char '*'
-  lift (skipMany spacenonewline)
+  lift skipNonNewlineSpaces
   -- this amount can have a price; balance assertions ignore it,
   -- but balance assignments will use it
   a <- amountpnolotprices <?> "amount (for a balance assertion or assignment)"
@@ -776,10 +784,10 @@ lotpricep :: JournalParser m ()
 lotpricep = label "ledger-style lot price" $ do
   char '{'
   doublebrace <- option False $ char '{' >> pure True
-  _fixed <- fmap isJust $ optional $ lift (skipMany spacenonewline) >> char '='
-  lift (skipMany spacenonewline)
+  _fixed <- fmap isJust $ optional $ lift skipNonNewlineSpaces >> char '='
+  lift skipNonNewlineSpaces
   _a <- amountwithoutpricep
-  lift (skipMany spacenonewline)
+  lift skipNonNewlineSpaces
   char '}'
   when (doublebrace) $ void $ char '}'
   return ()
@@ -789,9 +797,9 @@ lotpricep = label "ledger-style lot price" $ do
 lotdatep :: JournalParser m ()
 lotdatep = (do
   char '['
-  lift (skipMany spacenonewline)
+  lift skipNonNewlineSpaces
   _d <- datep
-  lift (skipMany spacenonewline)
+  lift skipNonNewlineSpaces
   char ']'
   return ()
   ) <?> "ledger-style lot date"
@@ -808,7 +816,7 @@ lotdatep = (do
 -- seen following the decimal mark), the decimal mark character used if any,
 -- and the digit group style if any.
 --
-numberp :: Maybe AmountStyle -> TextParser m (Quantity, Int, Maybe Char, Maybe DigitGroupStyle)
+numberp :: Maybe AmountStyle -> TextParser m (Quantity, Word8, Maybe Char, Maybe DigitGroupStyle)
 numberp suggestedStyle = label "number" $ do
     -- a number is an optional sign followed by a sequence of digits possibly
     -- interspersed with periods, commas, or both
@@ -816,13 +824,13 @@ numberp suggestedStyle = label "number" $ do
     sign <- signp
     rawNum <- either (disambiguateNumber suggestedStyle) id <$> rawnumberp
     mExp <- optional $ try $ exponentp
-    dbg8 "numberp suggestedStyle" suggestedStyle `seq` return ()
-    case dbg8 "numberp quantity,precision,mdecimalpoint,mgrps"
+    dbg7 "numberp suggestedStyle" suggestedStyle `seq` return ()
+    case dbg7 "numberp quantity,precision,mdecimalpoint,mgrps"
            $ fromRawNumber rawNum mExp of
       Left errMsg -> Fail.fail errMsg
       Right (q, p, d, g) -> pure (sign q, p, d, g)
 
-exponentp :: TextParser m Int
+exponentp :: TextParser m Integer
 exponentp = char' 'e' *> signp <*> decimal <?> "exponent"
 
 -- | Interpret a raw number as a decimal number.
@@ -834,50 +842,40 @@ exponentp = char' 'e' *> signp <*> decimal <?> "exponent"
 -- - the digit group style, if any (digit group character and sizes of digit groups)
 fromRawNumber
   :: RawNumber
-  -> Maybe Int
+  -> Maybe Integer
   -> Either String
-            (Quantity, Int, Maybe Char, Maybe DigitGroupStyle)
-fromRawNumber raw mExp = case raw of
-
-  NoSeparators digitGrp mDecimals ->
-    let mDecPt = fmap fst mDecimals
-        decimalGrp = maybe mempty snd mDecimals
-
-        (quantity, precision) =
-          maybe id applyExp mExp $ toQuantity digitGrp decimalGrp
-
-    in  Right (quantity, precision, mDecPt, Nothing)
-
-  WithSeparators digitSep digitGrps mDecimals -> case mExp of
-    Nothing ->
-      let mDecPt = fmap fst mDecimals
-          decimalGrp = maybe mempty snd mDecimals
-          digitGroupStyle = DigitGroups digitSep (groupSizes digitGrps)
-
-          (quantity, precision) = toQuantity (mconcat digitGrps) decimalGrp
-
-      in  Right (quantity, precision, mDecPt, Just digitGroupStyle)
-    Just _ -> Left
-      "invalid number: mixing digit separators with exponents is not allowed"
-
+            (Quantity, Word8, Maybe Char, Maybe DigitGroupStyle)
+fromRawNumber (WithSeparators _ _ _) (Just _) =
+    Left "invalid number: mixing digit separators with exponents is not allowed"
+fromRawNumber raw mExp = do
+    (quantity, precision) <- toQuantity (fromMaybe 0 mExp) (digitGroup raw) (decimalGroup raw)
+    return (quantity, precision, mDecPt raw, digitGroupStyle raw)
   where
+    toQuantity :: Integer -> DigitGrp -> DigitGrp -> Either String (Quantity, Word8)
+    toQuantity e preDecimalGrp postDecimalGrp
+      | precision < 0   = Right (Decimal 0 (digitGrpNum * 10^(-precision)), 0)
+      | precision < 256 = Right (Decimal precision8 digitGrpNum, precision8)
+      | otherwise = Left "invalid number: numbers with more than 255 decimal digits are not allowed at this time"
+      where
+        digitGrpNum = digitGroupNumber $ preDecimalGrp <> postDecimalGrp
+        precision   = toInteger (digitGroupLength postDecimalGrp) - e
+        precision8  = fromIntegral precision :: Word8
+
+    mDecPt (NoSeparators _ mDecimals)           = fst <$> mDecimals
+    mDecPt (WithSeparators _ _ mDecimals)       = fst <$> mDecimals
+    decimalGroup (NoSeparators _ mDecimals)     = maybe mempty snd mDecimals
+    decimalGroup (WithSeparators _ _ mDecimals) = maybe mempty snd mDecimals
+    digitGroup (NoSeparators digitGrp _)        = digitGrp
+    digitGroup (WithSeparators _ digitGrps _)   = mconcat digitGrps
+    digitGroupStyle (NoSeparators _ _)          = Nothing
+    digitGroupStyle (WithSeparators sep grps _) = Just . DigitGroups sep $ groupSizes grps
+
     -- Outputs digit group sizes from least significant to most significant
-    groupSizes :: [DigitGrp] -> [Int]
-    groupSizes digitGrps = reverse $ case map digitGroupLength digitGrps of
+    groupSizes :: [DigitGrp] -> [Word8]
+    groupSizes digitGrps = reverse $ case map (fromIntegral . digitGroupLength) digitGrps of
       (a:b:cs) | a < b -> b:cs
       gs               -> gs
 
-    toQuantity :: DigitGrp -> DigitGrp -> (Quantity, Int)
-    toQuantity preDecimalGrp postDecimalGrp = (quantity, precision)
-      where
-        quantity = Decimal (fromIntegral precision)
-                           (digitGroupNumber $ preDecimalGrp <> postDecimalGrp)
-        precision = digitGroupLength postDecimalGrp
-
-    applyExp :: Int -> (Decimal, Int) -> (Decimal, Int)
-    applyExp exponent (quantity, precision) =
-      (quantity * 10^^exponent, max 0 (precision - exponent))
-
 
 disambiguateNumber :: Maybe AmountStyle -> AmbiguousNumber -> RawNumber
 disambiguateNumber suggestedStyle (AmbiguousNumber grp1 sep grp2) =
@@ -892,7 +890,7 @@ disambiguateNumber suggestedStyle (AmbiguousNumber grp1 sep grp2) =
     isValidDecimalBy c = \case
       AmountStyle{asdecimalpoint = Just d} -> d == c
       AmountStyle{asdigitgroups = Just (DigitGroups g _)} -> g /= c
-      AmountStyle{asprecision = 0} -> False
+      AmountStyle{asprecision = Precision 0} -> False
       _ -> True
 
 -- | Parse and interpret the structure of a number without external hints.
@@ -930,7 +928,7 @@ rawnumberp = label "number" $ do
                   parseErrorAt off "invalid number (excessive trailing digits)"
     Nothing -> pure ()
 
-  return $ dbg8 "rawnumberp" rawNumber
+  return $ dbg7 "rawnumberp" rawNumber
   where
 
   leadingDecimalPt :: TextParser m RawNumber
@@ -1003,17 +1001,17 @@ data AmbiguousNumber = AmbiguousNumber DigitGrp Char DigitGrp
 -- | Description of a single digit group in a number literal.
 -- "Thousands" is one well known digit grouping, but there are others.
 data DigitGrp = DigitGrp {
-  digitGroupLength :: !Int,    -- ^ The number of digits in this group.
-  digitGroupNumber :: !Integer -- ^ The natural number formed by this group's digits.
+  digitGroupLength :: !Word,    -- ^ The number of digits in this group.
+                                -- This is Word to avoid the need to do overflow
+                                -- checking for the Semigroup instance of DigitGrp.
+  digitGroupNumber :: !Integer  -- ^ The natural number formed by this group's digits. This should always be positive.
 } deriving (Eq)
 
 -- | A custom show instance, showing digit groups as the parser saw them.
 instance Show DigitGrp where
-  show (DigitGrp len num)
-    | len > 0 = "\"" ++ padding ++ numStr ++ "\""
-    | otherwise = "\"\""
+  show (DigitGrp len num) = "\"" ++ padding ++ numStr ++ "\""
     where numStr = show num
-          padding = replicate (len - length numStr) '0'
+          padding = genericReplicate (toInteger len - toInteger (length numStr)) '0'
 
 instance Sem.Semigroup DigitGrp where
   DigitGrp l1 n1 <> DigitGrp l2 n2 = DigitGrp (l1 + l2) (n1 * 10^l2 + n2)
@@ -1037,7 +1035,7 @@ multilinecommentp = startComment *> anyLine `skipManyTill` endComment
     startComment = string "comment" *> trailingSpaces
     endComment = eof <|> string "end comment" *> trailingSpaces
 
-    trailingSpaces = skipMany spacenonewline <* newline
+    trailingSpaces = skipNonNewlineSpaces <* newline
     anyLine = void $ takeWhileP Nothing (\c -> c /= '\n') *> newline
 
 {-# INLINABLE multilinecommentp #-}
@@ -1047,7 +1045,7 @@ multilinecommentp = startComment *> anyLine `skipManyTill` endComment
 -- is semicolon, hash, or star.
 emptyorcommentlinep :: TextParser m ()
 emptyorcommentlinep = do
-  skipMany spacenonewline
+  skipNonNewlineSpaces
   skiplinecommentp <|> void newline
   where
     skiplinecommentp :: TextParser m ()
@@ -1076,13 +1074,13 @@ emptyorcommentlinep = do
 --
 followingcommentp' :: (Monoid a, Show a) => TextParser m a -> TextParser m (Text, a)
 followingcommentp' contentp = do
-  skipMany spacenonewline
+  skipNonNewlineSpaces
   -- there can be 0 or 1 sameLine
   sameLine <- try headerp *> ((:[]) <$> match' contentp) <|> pure []
   _ <- eolof
   -- there can be 0 or more nextLines
   nextLines <- many $
-    try (skipSome spacenonewline *> headerp) *> match' contentp <* eolof
+    try (skipNonNewlineSpaces1 *> headerp) *> match' contentp <* eolof
   let
     -- if there's just a next-line comment, insert an empty same-line comment
     -- so the next-line comment doesn't get rendered as a same-line comment.
@@ -1094,7 +1092,7 @@ followingcommentp' contentp = do
   pure (strippedCommentText, commentContent)
 
   where
-    headerp = char ';' *> skipMany spacenonewline
+    headerp = char ';' *> skipNonNewlineSpaces
 
 {-# INLINABLE followingcommentp' #-}
 
@@ -1158,7 +1156,7 @@ commenttagsp = do
       if T.null name
         then commenttagsp
         else do
-          skipMany spacenonewline
+          skipNonNewlineSpaces
           val <- tagValue
           let tag = (name, val)
           (tag:) <$> commenttagsp
@@ -1256,7 +1254,7 @@ commenttagsanddatesp mYear = do
 
     atColon :: Text -> TextParser m ([Tag], [DateTag])
     atColon name = char ':' *> do
-      skipMany spacenonewline
+      skipNonNewlineSpaces
       (tags, dateTags) <- case name of
         ""      -> pure ([], [])
         "date"  -> dateValue name
@@ -1342,38 +1340,38 @@ tests_Common = tests "Common" [
 
    tests "amountp" [
     test "basic"                  $ assertParseEq amountp "$47.18"     (usd 47.18)
-   ,test "ends with decimal mark" $ assertParseEq amountp "$1."        (usd 1  `withPrecision` 0)
+   ,test "ends with decimal mark" $ assertParseEq amountp "$1."        (usd 1  `withPrecision` Precision 0)
    ,test "unit price"             $ assertParseEq amountp "$10 @ €0.5"
       -- not precise enough:
       -- (usd 10 `withPrecision` 0 `at` (eur 0.5 `withPrecision` 1)) -- `withStyle` asdecimalpoint=Just '.'
       amount{
          acommodity="$"
         ,aquantity=10 -- need to test internal precision with roundTo ? I think not
-        ,astyle=amountstyle{asprecision=0, asdecimalpoint=Nothing}
+        ,astyle=amountstyle{asprecision=Precision 0, asdecimalpoint=Nothing}
         ,aprice=Just $ UnitPrice $
           amount{
              acommodity="€"
             ,aquantity=0.5
-            ,astyle=amountstyle{asprecision=1, asdecimalpoint=Just '.'}
+            ,astyle=amountstyle{asprecision=Precision 1, asdecimalpoint=Just '.'}
             }
         }
    ,test "total price"            $ assertParseEq amountp "$10 @@ €5"
       amount{
          acommodity="$"
         ,aquantity=10
-        ,astyle=amountstyle{asprecision=0, asdecimalpoint=Nothing}
+        ,astyle=amountstyle{asprecision=Precision 0, asdecimalpoint=Nothing}
         ,aprice=Just $ TotalPrice $
           amount{
              acommodity="€"
             ,aquantity=5
-            ,astyle=amountstyle{asprecision=0, asdecimalpoint=Nothing}
+            ,astyle=amountstyle{asprecision=Precision 0, asdecimalpoint=Nothing}
             }
         }
    ,test "unit price, parenthesised" $ assertParse amountp "$10 (@) €0.5"
    ,test "total price, parenthesised" $ assertParse amountp "$10 (@@) €0.5"
    ]
 
-  ,let p = lift (numberp Nothing) :: JournalParser IO (Quantity, Int, Maybe Char, Maybe DigitGroupStyle) in
+  ,let p = lift (numberp Nothing) :: JournalParser IO (Quantity, Word8, Maybe Char, Maybe DigitGroupStyle) in
    test "numberp" $ do
      assertParseEq p "0"          (0, 0, Nothing, Nothing)
      assertParseEq p "1"          (1, 0, Nothing, Nothing)
@@ -1393,6 +1391,8 @@ tests_Common = tests "Common" [
      assertParseError p "1..1" ""
      assertParseError p ".1," ""
      assertParseError p ",1." ""
+     assertParseEq    p "1.555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555" (1.555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555, 255, Just '.', Nothing)
+     assertParseError p "1.5555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555" ""
 
   ,tests "spaceandamountormissingp" [
      test "space and amount" $ assertParseEq spaceandamountormissingp " $47.18" (Mixed [usd 47.18])
