@@ -4,7 +4,6 @@ Postings report, used by the register command.
 
 -}
 
-{-# LANGUAGE DeriveDataTypeable #-}
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE RecordWildCards #-}
@@ -72,11 +71,11 @@ postingsReport ropts@ReportOpts{..} q j =
     where
       reportspan  = adjustReportDates ropts q j
       whichdate   = whichDateFromOpts ropts
-      depth       = queryDepth q
+      mdepth      = queryDepth q
       styles      = journalCommodityStyles j
       priceoracle = journalPriceOracle infer_value_ j
       multiperiod = interval_ /= NoInterval
-      today       = fromMaybe (error' "postingsReport: could not pick a valuation date, ReportOpts today_ is unset") today_
+      today       = fromMaybe (error' "postingsReport: could not pick a valuation date, ReportOpts today_ is unset") today_  -- PARTIAL:
 
       -- postings to be included in the report, and similarly-matched postings before the report start date
       (precedingps, reportps) = matchedPostingsBeforeAndDuring ropts q j reportspan
@@ -84,7 +83,7 @@ postingsReport ropts@ReportOpts{..} q j =
       -- Postings, or summary postings with their subperiod's end date, to be displayed.
       displayps :: [(Posting, Maybe Day)]
         | multiperiod =
-            let summaryps = summarisePostingsByInterval interval_ whichdate depth showempty reportspan reportps
+            let summaryps = summarisePostingsByInterval interval_ whichdate mdepth showempty reportspan reportps
             in [(pvalue p lastday, Just periodend) | (p, periodend) <- summaryps, let lastday = addDays (-1) periodend]
         | otherwise =
             [(pvalue p reportorjournallast, Nothing) | p <- reportps]
@@ -95,13 +94,13 @@ postingsReport ropts@ReportOpts{..} q j =
             where
               mreportlast = reportPeriodLastDay ropts
           reportorjournallast =
-            fromMaybe (error' "postingsReport: expected a non-empty journal") $ -- XXX shouldn't happen
+            fromMaybe (error' "postingsReport: expected a non-empty journal") $  -- PARTIAL: shouldn't happen
             reportPeriodOrJournalLastDay ropts j
 
       -- Posting report items ready for display.
       items =
-        dbg1 "postingsReport items" $
-        postingsReportItems displayps (nullposting,Nothing) whichdate depth startbal runningcalc startnum
+        dbg4 "postingsReport items" $
+        postingsReportItems displayps (nullposting,Nothing) whichdate mdepth startbal runningcalc startnum
         where
           -- In historical mode we'll need a starting balance, which we
           -- may be converting to value per hledger_options.m4.md "Effect
@@ -118,7 +117,7 @@ postingsReport ropts@ReportOpts{..} q j =
                   -- XXX constrain valuation type to AtDate daybeforereportstart here ?
                 where
                   daybeforereportstart =
-                    maybe (error' "postingsReport: expected a non-empty journal") -- XXX shouldn't happen
+                    maybe (error' "postingsReport: expected a non-empty journal")  -- PARTIAL: shouldn't happen
                     (addDays (-1))
                     $ reportPeriodOrJournalStart ropts j
 
@@ -145,16 +144,16 @@ adjustReportDates :: ReportOpts -> Query -> Journal -> DateSpan
 adjustReportDates opts q j = reportspan
   where
     -- see also multiBalanceReport
-    requestedspan       = dbg1 "requestedspan"       $ queryDateSpan' q                                       -- span specified by -b/-e/-p options and query args
-    journalspan         = dbg1 "journalspan"         $ dates `spanUnion` date2s                               -- earliest and latest dates (or date2s) in the journal
+    requestedspan       = dbg3 "requestedspan"       $ queryDateSpan' q                                       -- span specified by -b/-e/-p options and query args
+    journalspan         = dbg3 "journalspan"         $ dates `spanUnion` date2s                               -- earliest and latest dates (or date2s) in the journal
       where
         dates  = journalDateSpan False j
         date2s = journalDateSpan True  j
-    requestedspanclosed = dbg1 "requestedspanclosed" $ requestedspan `spanDefaultsFrom` journalspan           -- if open-ended, close it using the journal's dates (if any)
-    intervalspans       = dbg1 "intervalspans"       $ splitSpan (interval_ opts) requestedspanclosed  -- get the whole intervals enclosing that
-    mreportstart        = dbg1 "reportstart"         $ maybe Nothing spanStart $ headMay intervalspans        -- start of the first interval, or open ended
-    mreportend          = dbg1 "reportend"           $ maybe Nothing spanEnd   $ lastMay intervalspans        -- end of the last interval, or open ended
-    reportspan          = dbg1 "reportspan"          $ DateSpan mreportstart mreportend                       -- the requested span enlarged to whole intervals if possible
+    requestedspanclosed = dbg3 "requestedspanclosed" $ requestedspan `spanDefaultsFrom` journalspan           -- if open-ended, close it using the journal's dates (if any)
+    intervalspans       = dbg3 "intervalspans"       $ splitSpan (interval_ opts) requestedspanclosed  -- get the whole intervals enclosing that
+    mreportstart        = dbg3 "reportstart"         $ maybe Nothing spanStart $ headMay intervalspans        -- start of the first interval, or open ended
+    mreportend          = dbg3 "reportend"           $ maybe Nothing spanEnd   $ lastMay intervalspans        -- end of the last interval, or open ended
+    reportspan          = dbg3 "reportspan"          $ DateSpan mreportstart mreportend                       -- the requested span enlarged to whole intervals if possible
 
 -- | Find postings matching a given query, within a given date span,
 -- and also any similarly-matched postings before that date span.
@@ -162,45 +161,46 @@ adjustReportDates opts q j = reportspan
 -- A helper for the postings report.
 matchedPostingsBeforeAndDuring :: ReportOpts -> Query -> Journal -> DateSpan -> ([Posting],[Posting])
 matchedPostingsBeforeAndDuring opts q j (DateSpan mstart mend) =
-  dbg1 "beforeps, duringps" $ span (beforestartq `matchesPosting`) beforeandduringps
+  dbg5 "beforeps, duringps" $ span (beforestartq `matchesPosting`) beforeandduringps
   where
-    beforestartq = dbg1 "beforestartq" $ dateqtype $ DateSpan Nothing mstart
+    beforestartq = dbg3 "beforestartq" $ dateqtype $ DateSpan Nothing mstart
     beforeandduringps =
-      dbg1 "ps5" $ sortOn sortdate $                                           -- sort postings by date or date2
-      dbg1 "ps4" $ (if invert_ opts then map negatePostingAmount else id) $    -- with --invert, invert amounts
-      dbg1 "ps3" $ map (filterPostingAmount symq) $                            -- remove amount parts which the query's cur: terms would exclude
-      dbg1 "ps2" $ (if related_ opts then concatMap relatedPostings else id) $ -- with -r, replace each with its sibling postings
-      dbg1 "ps1" $ filter (beforeandduringq `matchesPosting`) $                -- filter postings by the query, with no start date or depth limit
+      dbg5 "ps5" $ sortOn sortdate $                                           -- sort postings by date or date2
+      dbg5 "ps4" $ (if invert_ opts then map negatePostingAmount else id) $    -- with --invert, invert amounts
+      dbg5 "ps3" $ map (filterPostingAmount symq) $                            -- remove amount parts which the query's cur: terms would exclude
+      dbg5 "ps2" $ (if related_ opts then concatMap relatedPostings else id) $ -- with -r, replace each with its sibling postings
+      dbg5 "ps1" $ filter (beforeandduringq `matchesPosting`) $                -- filter postings by the query, with no start date or depth limit
                   journalPostings $
                   journalSelectingAmountFromOpts opts j    -- maybe convert to cost early, will be seen by amt:. XXX what about converting to value ?
       where
-        beforeandduringq = dbg1 "beforeandduringq" $ And [depthless $ dateless q, beforeendq]
+        beforeandduringq = dbg4 "beforeandduringq" $ And [depthless $ dateless q, beforeendq]
           where
             depthless  = filterQuery (not . queryIsDepth)
             dateless   = filterQuery (not . queryIsDateOrDate2)
             beforeendq = dateqtype $ DateSpan Nothing mend
         sortdate = if date2_ opts then postingDate2 else postingDate
-        symq = dbg1 "symq" $ filterQuery queryIsSym q
+        symq = dbg4 "symq" $ filterQuery queryIsSym q
     dateqtype
       | queryIsDate2 dateq || (queryIsDate dateq && date2_ opts) = Date2
       | otherwise = Date
       where
-        dateq = dbg1 "dateq" $ filterQuery queryIsDateOrDate2 $ dbg1 "q" q  -- XXX confused by multiple date:/date2: ?
+        dateq = dbg4 "dateq" $ filterQuery queryIsDateOrDate2 $ dbg4 "q" q  -- XXX confused by multiple date:/date2: ?
 
 -- | Generate postings report line items from a list of postings or (with
 -- non-Nothing dates attached) summary postings.
-postingsReportItems :: [(Posting,Maybe Day)] -> (Posting,Maybe Day) -> WhichDate -> Int -> MixedAmount -> (Int -> MixedAmount -> MixedAmount -> MixedAmount) -> Int -> [PostingsReportItem]
+postingsReportItems :: [(Posting,Maybe Day)] -> (Posting,Maybe Day) -> WhichDate -> Maybe Int -> MixedAmount -> (Int -> MixedAmount -> MixedAmount -> MixedAmount) -> Int -> [PostingsReportItem]
 postingsReportItems [] _ _ _ _ _ _ = []
-postingsReportItems ((p,menddate):ps) (pprev,menddateprev) wd d b runningcalcfn itemnum = i:(postingsReportItems ps (p,menddate) wd d b' runningcalcfn (itemnum+1))
-    where
-      i = mkpostingsReportItem showdate showdesc wd menddate p' b'
-      (showdate, showdesc) | isJust menddate = (menddate /= menddateprev,        False)
-                           | otherwise       = (isfirstintxn || isdifferentdate, isfirstintxn)
-      isfirstintxn = ptransaction p /= ptransaction pprev
-      isdifferentdate = case wd of PrimaryDate   -> postingDate p  /= postingDate pprev
-                                   SecondaryDate -> postingDate2 p /= postingDate2 pprev
-      p' = p{paccount= clipOrEllipsifyAccountName d $ paccount p}
-      b' = runningcalcfn itemnum b (pamount p)
+postingsReportItems ((p,menddate):ps) (pprev,menddateprev) wd d b runningcalcfn itemnum =
+    i:(postingsReportItems ps (p,menddate) wd d b' runningcalcfn (itemnum+1))
+  where
+    i = mkpostingsReportItem showdate showdesc wd menddate p' b'
+    (showdate, showdesc) | isJust menddate = (menddate /= menddateprev,        False)
+                         | otherwise       = (isfirstintxn || isdifferentdate, isfirstintxn)
+    isfirstintxn = ptransaction p /= ptransaction pprev
+    isdifferentdate = case wd of PrimaryDate   -> postingDate p  /= postingDate pprev
+                                 SecondaryDate -> postingDate2 p /= postingDate2 pprev
+    p' = p{paccount= clipOrEllipsifyAccountName d $ paccount p}
+    b' = runningcalcfn itemnum b (pamount p)
 
 -- | Generate one postings report line item, containing the posting,
 -- the current running balance, and optionally the posting date and/or
@@ -221,11 +221,11 @@ mkpostingsReportItem showdate showdesc wd menddate p b =
 -- | Convert a list of postings into summary postings, one per interval,
 -- aggregated to the specified depth if any.
 -- Each summary posting will have a non-Nothing interval end date.
-summarisePostingsByInterval :: Interval -> WhichDate -> Int -> Bool -> DateSpan -> [Posting] -> [SummaryPosting]
-summarisePostingsByInterval interval wd depth showempty reportspan ps = concatMap summarisespan $ splitSpan interval reportspan
-    where
-      summarisespan s = summarisePostingsInDateSpan s wd depth showempty (postingsinspan s)
-      postingsinspan s = filter (isPostingInDateSpan' wd s) ps
+summarisePostingsByInterval :: Interval -> WhichDate -> Maybe Int -> Bool -> DateSpan -> [Posting] -> [SummaryPosting]
+summarisePostingsByInterval interval wd mdepth showempty reportspan ps = concatMap summarisespan $ splitSpan interval reportspan
+  where
+    summarisespan s = summarisePostingsInDateSpan s wd mdepth showempty (postingsinspan s)
+    postingsinspan s = filter (isPostingInDateSpan' wd s) ps
 
 -- | Given a date span (representing a report interval) and a list of
 -- postings within it, aggregate the postings into one summary posting per
@@ -239,28 +239,27 @@ summarisePostingsByInterval interval wd depth showempty reportspan ps = concatMa
 -- The showempty flag includes spans with no postings and also postings
 -- with 0 amount.
 --
-summarisePostingsInDateSpan :: DateSpan -> WhichDate -> Int -> Bool -> [Posting] -> [SummaryPosting]
-summarisePostingsInDateSpan (DateSpan b e) wd depth showempty ps
-    | null ps && (isNothing b || isNothing e) = []
-    | null ps && showempty = [(summaryp, e')]
-    | otherwise = summarypes
-    where
-      postingdate = if wd == PrimaryDate then postingDate else postingDate2
-      b' = fromMaybe (maybe nulldate postingdate $ headMay ps) b
-      e' = fromMaybe (maybe (addDays 1 nulldate) postingdate $ lastMay ps) e
-      summaryp = nullposting{pdate=Just b'}
-      clippedanames | depth > 0 = nub $ map (clipAccountName depth) anames
-                    | otherwise = ["..."]
-      summaryps | depth > 0 = [summaryp{paccount=a,pamount=balance a} | a <- clippedanames]
-                | otherwise = [summaryp{paccount="...",pamount=sum $ map pamount ps}]
-      summarypes = map (, e') $ (if showempty then id else filter (not . mixedAmountLooksZero . pamount)) summaryps
-      anames = nubSort $ map paccount ps
-      -- aggregate balances by account, like ledgerFromJournal, then do depth-clipping
-      accts = accountsFromPostings ps
-      balance a = maybe nullmixedamt bal $ lookupAccount a accts
-        where
-          bal = if isclipped a then aibalance else aebalance
-          isclipped a = accountNameLevel a >= depth
+summarisePostingsInDateSpan :: DateSpan -> WhichDate -> Maybe Int -> Bool -> [Posting] -> [SummaryPosting]
+summarisePostingsInDateSpan (DateSpan b e) wd mdepth showempty ps
+  | null ps && (isNothing b || isNothing e) = []
+  | null ps && showempty = [(summaryp, e')]
+  | otherwise = summarypes
+  where
+    postingdate = if wd == PrimaryDate then postingDate else postingDate2
+    b' = fromMaybe (maybe nulldate postingdate $ headMay ps) b
+    e' = fromMaybe (maybe (addDays 1 nulldate) postingdate $ lastMay ps) e
+    summaryp = nullposting{pdate=Just b'}
+    clippedanames = nub $ map (clipAccountName mdepth) anames
+    summaryps | mdepth == Just 0 = [summaryp{paccount="...",pamount=sum $ map pamount ps}]
+              | otherwise        = [summaryp{paccount=a,pamount=balance a} | a <- clippedanames]
+    summarypes = map (, e') $ (if showempty then id else filter (not . mixedAmountLooksZero . pamount)) summaryps
+    anames = nubSort $ map paccount ps
+    -- aggregate balances by account, like ledgerFromJournal, then do depth-clipping
+    accts = accountsFromPostings ps
+    balance a = maybe nullmixedamt bal $ lookupAccount a accts
+      where
+        bal = if isclipped a then aibalance else aebalance
+        isclipped a = maybe True (accountNameLevel a >=) mdepth
 
 negatePostingAmount :: Posting -> Posting
 negatePostingAmount p = p { pamount = negate $ pamount p }
@@ -277,16 +276,16 @@ tests_PostingsReport = tests "PostingsReport" [
     (Any, samplejournal) `gives` 13
     -- register --depth just clips account names
     (Depth 2, samplejournal) `gives` 13
-    (And [Depth 1, StatusQ Cleared, Acct "expenses"], samplejournal) `gives` 2
-    (And [And [Depth 1, StatusQ Cleared], Acct "expenses"], samplejournal) `gives` 2
+    (And [Depth 1, StatusQ Cleared, Acct (toRegex' "expenses")], samplejournal) `gives` 2
+    (And [And [Depth 1, StatusQ Cleared], Acct (toRegex' "expenses")], samplejournal) `gives` 2
     -- with query and/or command-line options
     (length $ snd $ postingsReport defreportopts Any samplejournal) @?= 13
     (length $ snd $ postingsReport defreportopts{interval_=Months 1} Any samplejournal) @?= 11
     (length $ snd $ postingsReport defreportopts{interval_=Months 1, empty_=True} Any samplejournal) @?= 20
-    (length $ snd $ postingsReport defreportopts (Acct "assets:bank:checking") samplejournal) @?= 5
+    (length $ snd $ postingsReport defreportopts (Acct (toRegex' "assets:bank:checking")) samplejournal) @?= 5
 
      -- (defreportopts, And [Acct "a a", Acct "'b"], samplejournal2) `gives` 0
-     -- [(Just (parsedate "2008-01-01","income"),assets:bank:checking             $1,$1)
+     -- [(Just (fromGregorian 2008 01 01,"income"),assets:bank:checking             $1,$1)
      -- ,(Nothing,income:salary                   $-1,0)
      -- ,(Just (2008-06-01,"gift"),assets:bank:checking             $1,$1)
      -- ,(Nothing,income:gifts                    $-1,0)
@@ -374,7 +373,7 @@ tests_PostingsReport = tests "PostingsReport" [
         j <- samplejournal
         let gives displayexpr =
                 (registerdates (postingsReportAsText opts $ postingsReport opts (queryFromOpts date1 opts) j) `is`)
-                    where opts = defreportopts{display_=Just displayexpr}
+                    where opts = defreportopts
         "d<[2008/6/2]"  `gives` ["2008/01/01","2008/06/01"]
         "d<=[2008/6/2]" `gives` ["2008/01/01","2008/06/01","2008/06/02"]
         "d=[2008/6/2]"  `gives` ["2008/06/02"]
@@ -432,12 +431,12 @@ tests_PostingsReport = tests "PostingsReport" [
     -}
 
   ,test "summarisePostingsByInterval" $
-    summarisePostingsByInterval (Quarters 1) PrimaryDate 99999 False (DateSpan Nothing Nothing) [] @?= []
+    summarisePostingsByInterval (Quarters 1) PrimaryDate Nothing False (DateSpan Nothing Nothing) [] @?= []
 
   -- ,tests_summarisePostingsInDateSpan = [
     --  "summarisePostingsInDateSpan" ~: do
     --   let gives (b,e,depth,showempty,ps) =
-    --           (summarisePostingsInDateSpan (mkdatespan b e) depth showempty ps `is`)
+    --           (summarisePostingsInDateSpan (DateSpan b e) depth showempty ps `is`)
     --   let ps =
     --           [
     --            nullposting{lpdescription="desc",lpaccount="expenses:food:groceries",lpamount=Mixed [usd 1]}
@@ -449,25 +448,25 @@ tests_PostingsReport = tests "PostingsReport" [
     --    []
     --   ("2008/01/01","2009/01/01",0,9999,True,[]) `gives`
     --    [
-    --     nullposting{lpdate=parsedate "2008/01/01",lpdescription="- 2008/12/31"}
+    --     nullposting{lpdate=fromGregorian 2008 01 01,lpdescription="- 2008/12/31"}
     --    ]
     --   ("2008/01/01","2009/01/01",0,9999,False,ts) `gives`
     --    [
-    --     nullposting{lpdate=parsedate "2008/01/01",lpdescription="- 2008/12/31",lpaccount="expenses:food",          lpamount=Mixed [usd 4]}
-    --    ,nullposting{lpdate=parsedate "2008/01/01",lpdescription="- 2008/12/31",lpaccount="expenses:food:dining",   lpamount=Mixed [usd 10]}
-    --    ,nullposting{lpdate=parsedate "2008/01/01",lpdescription="- 2008/12/31",lpaccount="expenses:food:groceries",lpamount=Mixed [usd 1]}
+    --     nullposting{lpdate=fromGregorian 2008 01 01,lpdescription="- 2008/12/31",lpaccount="expenses:food",          lpamount=Mixed [usd 4]}
+    --    ,nullposting{lpdate=fromGregorian 2008 01 01,lpdescription="- 2008/12/31",lpaccount="expenses:food:dining",   lpamount=Mixed [usd 10]}
+    --    ,nullposting{lpdate=fromGregorian 2008 01 01,lpdescription="- 2008/12/31",lpaccount="expenses:food:groceries",lpamount=Mixed [usd 1]}
     --    ]
     --   ("2008/01/01","2009/01/01",0,2,False,ts) `gives`
     --    [
-    --     nullposting{lpdate=parsedate "2008/01/01",lpdescription="- 2008/12/31",lpaccount="expenses:food",lpamount=Mixed [usd 15]}
+    --     nullposting{lpdate=fromGregorian 2008 01 01,lpdescription="- 2008/12/31",lpaccount="expenses:food",lpamount=Mixed [usd 15]}
     --    ]
     --   ("2008/01/01","2009/01/01",0,1,False,ts) `gives`
     --    [
-    --     nullposting{lpdate=parsedate "2008/01/01",lpdescription="- 2008/12/31",lpaccount="expenses",lpamount=Mixed [usd 15]}
+    --     nullposting{lpdate=fromGregorian 2008 01 01,lpdescription="- 2008/12/31",lpaccount="expenses",lpamount=Mixed [usd 15]}
     --    ]
     --   ("2008/01/01","2009/01/01",0,0,False,ts) `gives`
     --    [
-    --     nullposting{lpdate=parsedate "2008/01/01",lpdescription="- 2008/12/31",lpaccount="",lpamount=Mixed [usd 15]}
+    --     nullposting{lpdate=fromGregorian 2008 01 01,lpdescription="- 2008/12/31",lpaccount="",lpamount=Mixed [usd 15]}
     --    ]
 
  ]
