@@ -4,7 +4,9 @@ Options common to most hledger reports.
 
 -}
 
-{-# LANGUAGE OverloadedStrings, RecordWildCards, LambdaCase, DeriveDataTypeable #-}
+{-# LANGUAGE LambdaCase #-}
+{-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE RecordWildCards #-}
 
 module Hledger.Reports.ReportOptions (
   ReportOpts(..),
@@ -46,15 +48,15 @@ module Hledger.Reports.ReportOptions (
 where
 
 import Control.Applicative ((<|>))
-import Data.Data (Data)
 import Data.List.Extra (nubSort)
-import Data.Maybe
+import Data.Maybe (fromMaybe, isJust)
 import qualified Data.Text as T
-import Data.Typeable (Typeable)
-import Data.Time.Calendar
-import Data.Default
-import Safe
-import System.Console.ANSI (hSupportsANSI)
+import Data.Time.Calendar (Day, addDays, fromGregorian)
+import Data.Default (Default(..))
+import Safe (lastDef, lastMay)
+
+import System.Console.ANSI (hSupportsANSIColor)
+import System.Environment (lookupEnv)
 import System.IO (stdout)
 import Text.Megaparsec.Custom
 
@@ -71,20 +73,21 @@ data BalanceType = PeriodChange      -- ^ The change of balance in each period.
                  | HistoricalBalance -- ^ The historical ending balance, including the effect of
                                      --   all postings before the report period. Unless altered by,
                                      --   a query, this is what you would see on a bank statement.
-  deriving (Eq,Show,Data,Typeable)
+  deriving (Eq,Show)
 
 instance Default BalanceType where def = PeriodChange
 
 -- | Should accounts be displayed: in the command's default style, hierarchically, or as a flat list ?
-data AccountListMode = ALDefault | ALTree | ALFlat deriving (Eq, Show, Data, Typeable)
+data AccountListMode = ALFlat | ALTree deriving (Eq, Show)
 
-instance Default AccountListMode where def = ALDefault
+instance Default AccountListMode where def = ALFlat
 
 -- | Standard options for customising report filtering and output.
 -- Most of these correspond to standard hledger command-line options
 -- or query arguments, but not all. Some are used only by certain
 -- commands, as noted below.
 data ReportOpts = ReportOpts {
+     -- for most reports:
      today_          :: Maybe Day  -- ^ The current date. A late addition to ReportOpts.
                                    -- Optional, but when set it may affect some reports:
                                    -- Reports use it when picking a -V valuation date.
@@ -95,7 +98,6 @@ data ReportOpts = ReportOpts {
     ,value_          :: Maybe ValuationType  -- ^ What value should amounts be converted to ?
     ,infer_value_    :: Bool      -- ^ Infer market prices from transactions ?
     ,depth_          :: Maybe Int
-    ,display_        :: Maybe DisplayExp  -- XXX unused ?
     ,date2_          :: Bool
     ,empty_          :: Bool
     ,no_elide_       :: Bool
@@ -105,9 +107,11 @@ data ReportOpts = ReportOpts {
                                --   and quoted if needed (see 'quoteIfNeeded')
     --
     ,average_        :: Bool
-    -- register command only
+    -- for posting reports (register)
     ,related_        :: Bool
-    -- balance-type commands only
+    -- for account transactions reports (aregister)
+    ,txn_dates_      :: Bool
+    -- for balance reports (bal, bs, cf, is)
     ,balancetype_    :: BalanceType
     ,accountlistmode_ :: AccountListMode
     ,drop_           :: Int
@@ -126,9 +130,13 @@ data ReportOpts = ReportOpts {
       --   sign normalisation, converting normally negative subreports to
       --   normally positive for a more conventional display.
     ,color_          :: Bool
+      -- ^ Whether to use ANSI color codes in text output.
+      --   Influenced by the --color/colour flag (cf CliOptions),
+      --   whether stdout is an interactive terminal, and the value of
+      --   TERM and existence of NO_COLOR environment variables.
     ,forecast_       :: Maybe DateSpan
     ,transpose_      :: Bool
- } deriving (Show, Data, Typeable)
+ } deriving (Show)
 
 instance Default ReportOpts where def = defreportopts
 
@@ -168,7 +176,9 @@ rawOptsToReportOpts :: RawOpts -> IO ReportOpts
 rawOptsToReportOpts rawopts = checkReportOpts <$> do
   let rawopts' = checkRawOpts rawopts
   d <- getCurrentDay
-  color <- hSupportsANSI stdout
+  no_color <- isJust <$> lookupEnv "NO_COLOR"
+  supports_color <- hSupportsANSIColor stdout
+  let colorflag = stringopt "color" rawopts
   return defreportopts{
      today_       = Just d
     ,period_      = periodFromRawOpts d rawopts'
@@ -176,8 +186,7 @@ rawOptsToReportOpts rawopts = checkReportOpts <$> do
     ,statuses_    = statusesFromRawOpts rawopts'
     ,value_       = valuationTypeFromRawOpts rawopts'
     ,infer_value_ = boolopt "infer-value" rawopts'
-    ,depth_       = maybeintopt "depth" rawopts'
-    ,display_     = maybedisplayopt d rawopts'
+    ,depth_       = maybeposintopt "depth" rawopts'
     ,date2_       = boolopt "date2" rawopts'
     ,empty_       = boolopt "empty" rawopts'
     ,no_elide_    = boolopt "no-elide" rawopts'
@@ -186,16 +195,20 @@ rawOptsToReportOpts rawopts = checkReportOpts <$> do
     ,query_       = unwords . map quoteIfNeeded $ listofstringopt "args" rawopts' -- doesn't handle an arg like "" right
     ,average_     = boolopt "average" rawopts'
     ,related_     = boolopt "related" rawopts'
+    ,txn_dates_   = boolopt "txn-dates" rawopts'
     ,balancetype_ = balancetypeopt rawopts'
     ,accountlistmode_ = accountlistmodeopt rawopts'
-    ,drop_        = intopt "drop" rawopts'
+    ,drop_        = posintopt "drop" rawopts'
     ,row_total_   = boolopt "row-total" rawopts'
     ,no_total_    = boolopt "no-total" rawopts'
     ,sort_amount_ = boolopt "sort-amount" rawopts'
     ,percent_     = boolopt "percent" rawopts'
     ,invert_      = boolopt "invert" rawopts'
     ,pretty_tables_ = boolopt "pretty-tables" rawopts'
-    ,color_       = color
+    ,color_       = and [not no_color
+                        ,not $ colorflag `elem` ["never","no"]
+                        ,colorflag `elem` ["always","yes"] || supports_color
+                        ]
     ,forecast_    = forecastPeriodFromRawOpts d rawopts'
     ,transpose_   = boolopt "transpose" rawopts'
     }
@@ -225,7 +238,7 @@ checkReportOpts ropts@ReportOpts{..} =
 
 accountlistmodeopt :: RawOpts -> AccountListMode
 accountlistmodeopt =
-  fromMaybe ALDefault . choiceopt parse where
+  fromMaybe ALFlat . choiceopt parse where
     parse = \case
       "tree" -> Just ALTree
       "flat" -> Just ALFlat
@@ -308,7 +321,7 @@ intervalFromRawOpts = lastDef NoInterval . collectopts intervalfromrawopt
             (\e -> usageError $ "could not parse period option: "++customErrorBundlePretty e)
             extractIntervalOrNothing $
             parsePeriodExpr
-              (error' "intervalFromRawOpts: did not expect to need today's date here") -- should not happen; we are just getting the interval, which does not use the reference date
+              (error' "intervalFromRawOpts: did not expect to need today's date here")  -- PARTIAL: should not happen; we are just getting the interval, which does not use the reference date
               (stripquotes $ T.pack v)
       | n == "daily"     = Just $ Days 1
       | n == "weekly"    = Just $ Weeks 1
@@ -327,7 +340,7 @@ forecastPeriodFromRawOpts d opts =
     Just str ->
       either (\e -> usageError $ "could not parse forecast period : "++customErrorBundlePretty e) (Just . snd) $ 
       parsePeriodExpr d $ stripquotes $ T.pack str
-    
+
 -- | Extract the interval from the parsed -p/--period expression.
 -- Return Nothing if an interval is not explicitly defined.
 extractIntervalOrNothing :: (Interval, DateSpan) -> Maybe Interval
@@ -400,15 +413,6 @@ valuationTypeIsDefaultValue ropts =
     Just (AtDefault _) -> True
     _                  -> False
 
-type DisplayExp = String
-
-maybedisplayopt :: Day -> RawOpts -> Maybe DisplayExp
-maybedisplayopt d rawopts =
-    maybe Nothing (Just . regexReplaceBy "\\[.+?\\]" fixbracketeddatestr) $ maybestringopt "display" rawopts
-    where
-      fixbracketeddatestr "" = ""
-      fixbracketeddatestr s = "[" ++ fixSmartDateStr d (T.pack $ init $ tail s) ++ "]"
-
 -- | Select the Transaction date accessor based on --date2.
 transactionDateFn :: ReportOpts -> (Transaction -> Day)
 transactionDateFn ReportOpts{..} = if date2_ then transactionDate2 else tdate
@@ -423,10 +427,11 @@ whichDateFromOpts ReportOpts{..} = if date2_ then SecondaryDate else PrimaryDate
 
 -- | Legacy-compatible convenience aliases for accountlistmode_.
 tree_ :: ReportOpts -> Bool
-tree_ = (==ALTree) . accountlistmode_
+tree_ ReportOpts{accountlistmode_ = ALTree} = True
+tree_ ReportOpts{accountlistmode_ = ALFlat} = False
 
 flat_ :: ReportOpts -> Bool
-flat_ = (==ALFlat) . accountlistmode_
+flat_ = not . tree_
 
 -- depthFromOpts :: ReportOpts -> Int
 -- depthFromOpts opts = min (fromMaybe 99999 $ depth_ opts) (queryDepth $ queryFromOpts nulldate opts)
@@ -441,34 +446,30 @@ journalSelectingAmountFromOpts opts =
     _               -> id
 
 -- | Convert report options and arguments to a query.
+-- If there is a parsing problem, this function calls error.
 queryFromOpts :: Day -> ReportOpts -> Query
-queryFromOpts d ReportOpts{..} = simplifyQuery $ And $ [flagsq, argsq]
+queryFromOpts d ropts = simplifyQuery . And $ [flagsq, argsq]
   where
-    flagsq = And $
-              [(if date2_ then Date2 else Date) $ periodAsDateSpan period_]
-              ++ (if real_ then [Real True] else [])
-              ++ (if empty_ then [Empty True] else []) -- ?
-              ++ [Or $ map StatusQ statuses_]
-              ++ (maybe [] ((:[]) . Depth) depth_)
-    argsq = fst $ parseQuery d (T.pack query_)
+    flagsq = queryFromOptsOnly d ropts
+    argsq = fst $ either error' id $ parseQuery d (T.pack $ query_ ropts)  -- PARTIAL:
 
 -- | Convert report options to a query, ignoring any non-flag command line arguments.
 queryFromOptsOnly :: Day -> ReportOpts -> Query
-queryFromOptsOnly _d ReportOpts{..} = simplifyQuery flagsq
+queryFromOptsOnly _d ReportOpts{..} = simplifyQuery $ And flagsq
   where
-    flagsq = And $
-              [(if date2_ then Date2 else Date) $ periodAsDateSpan period_]
-              ++ (if real_ then [Real True] else [])
-              ++ (if empty_ then [Empty True] else []) -- ?
-              ++ [Or $ map StatusQ statuses_]
-              ++ (maybe [] ((:[]) . Depth) depth_)
+    flagsq = consIf   Real  real_
+           . consIf   Empty empty_
+           . consJust Depth depth_
+           $   [ (if date2_ then Date2 else Date) $ periodAsDateSpan period_
+               , Or $ map StatusQ statuses_
+               ]
+    consIf f b = if b then (f True:) else id
+    consJust f = maybe id ((:) . f)
 
 -- | Convert report options and arguments to query options.
+-- If there is a parsing problem, this function calls error.
 queryOptsFromOpts :: Day -> ReportOpts -> [QueryOpt]
-queryOptsFromOpts d ReportOpts{..} = flagsqopts ++ argsqopts
-  where
-    flagsqopts = []
-    argsqopts = snd $ parseQuery d (T.pack query_)
+queryOptsFromOpts d = snd . either error' id . parseQuery d . T.pack . query_  -- PARTIAL:
 
 -- Report dates.
 
@@ -557,17 +558,16 @@ reportPeriodOrJournalLastDay ropts j =
 tests_ReportOptions = tests "ReportOptions" [
    test "queryFromOpts" $ do
        queryFromOpts nulldate defreportopts @?= Any
-       queryFromOpts nulldate defreportopts{query_="a"} @?= Acct "a"
-       queryFromOpts nulldate defreportopts{query_="desc:'a a'"} @?= Desc "a a"
-       queryFromOpts nulldate defreportopts{period_=PeriodFrom (parsedate "2012/01/01"),query_="date:'to 2013'" }
-         @?= (Date $ mkdatespan "2012/01/01" "2013/01/01")
-       queryFromOpts nulldate defreportopts{query_="date2:'in 2012'"} @?= (Date2 $ mkdatespan "2012/01/01" "2013/01/01")
-       queryFromOpts nulldate defreportopts{query_="'a a' 'b"} @?= Or [Acct "a a", Acct "'b"]
+       queryFromOpts nulldate defreportopts{query_="a"} @?= Acct (toRegexCI' "a")
+       queryFromOpts nulldate defreportopts{query_="desc:'a a'"} @?= Desc (toRegexCI' "a a")
+       queryFromOpts nulldate defreportopts{period_=PeriodFrom (fromGregorian 2012 01 01),query_="date:'to 2013'" }
+         @?= (Date $ DateSpan (Just $ fromGregorian 2012 01 01) (Just $ fromGregorian 2013 01 01))
+       queryFromOpts nulldate defreportopts{query_="date2:'in 2012'"} @?= (Date2 $ DateSpan (Just $ fromGregorian 2012 01 01) (Just $ fromGregorian 2013 01 01))
+       queryFromOpts nulldate defreportopts{query_="'a a' 'b"} @?= Or [Acct $ toRegexCI' "a a", Acct $ toRegexCI' "'b"]
 
   ,test "queryOptsFromOpts" $ do
       queryOptsFromOpts nulldate defreportopts @?= []
       queryOptsFromOpts nulldate defreportopts{query_="a"} @?= []
-      queryOptsFromOpts nulldate defreportopts{period_=PeriodFrom (parsedate "2012/01/01")
+      queryOptsFromOpts nulldate defreportopts{period_=PeriodFrom (fromGregorian 2012 01 01)
                                               ,query_="date:'to 2013'"} @?= []
  ]
-
