@@ -1,6 +1,7 @@
 -- | String formatting helpers, starting to get a bit out of control.
 
 module Hledger.Utils.String (
+ takeEnd,
  -- * misc
  lowercase,
  uppercase,
@@ -12,7 +13,6 @@ module Hledger.Utils.String (
  singleQuoteIfNeeded,
  -- quotechars,
  -- whitespacechars,
- escapeQuotes,
  words',
  unwords',
  stripAnsi,
@@ -48,14 +48,22 @@ module Hledger.Utils.String (
  ) where
 
 
-import Data.Char
-import Data.List
-import Text.Megaparsec
-import Text.Megaparsec.Char
+import Data.Char (isDigit, isSpace, toLower, toUpper)
+import Data.List (intercalate, transpose)
+import Text.Megaparsec (Parsec, (<|>), (<?>), anySingle, between, many, noneOf,
+                        oneOf, parseMaybe, sepBy, takeWhileP, try)
+import Text.Megaparsec.Char (char, string)
 import Text.Printf (printf)
 
 import Hledger.Utils.Parse
-import Hledger.Utils.Regex
+
+
+-- | Take elements from the end of a list.
+takeEnd n l = go (drop n l) l
+  where
+    go (_:xs) (_:ys) = go xs ys
+    go []     r      = r
+    go _      []     = []
 
 lowercase, uppercase :: String -> String
 lowercase = map toLower
@@ -86,7 +94,7 @@ stripbrackets = dropWhile (`elem` "([") . reverse . dropWhile (`elem` "])") . re
 
 elideLeft :: Int -> String -> String
 elideLeft width s =
-    if length s > width then ".." ++ reverse (take (width - 2) $ reverse s) else s
+    if length s > width then ".." ++ takeEnd (width - 2) s else s
 
 elideRight :: Int -> String -> String
 elideRight width s =
@@ -111,8 +119,9 @@ underline s = s' ++ replicate (length s) '-' ++ "\n"
 -- | Double-quote this string if it contains whitespace, single quotes
 -- or double-quotes, escaping the quotes as needed.
 quoteIfNeeded :: String -> String
-quoteIfNeeded s | any (`elem` s) (quotechars++whitespacechars++redirectchars) = "\"" ++ escapeDoubleQuotes s ++ "\""
+quoteIfNeeded s | any (`elem` s) (quotechars++whitespacechars++redirectchars) = show s
                 | otherwise = s
+
 -- | Single-quote this string if it contains whitespace or double-quotes.
 -- No good for strings containing single quotes.
 singleQuoteIfNeeded :: String -> String
@@ -124,19 +133,13 @@ quotechars      = "'\""
 whitespacechars = " \t\n\r"
 redirectchars   = "<>"
 
-escapeDoubleQuotes :: String -> String
-escapeDoubleQuotes = regexReplace "\"" "\""
-
-escapeQuotes :: String -> String
-escapeQuotes = regexReplace "([\"'])" "\\1"
-
 -- | Quote-aware version of words - don't split on spaces which are inside quotes.
 -- NB correctly handles "a'b" but not "''a''". Can raise an error if parsing fails.
 words' :: String -> [String]
 words' "" = []
 words' s  = map stripquotes $ fromparse $ parsewithString p s
     where
-      p = do ss <- (singleQuotedPattern <|> doubleQuotedPattern <|> pattern) `sepBy` skipSome spacenonewline
+      p = do ss <- (singleQuotedPattern <|> doubleQuotedPattern <|> pattern) `sepBy` skipNonNewlineSpaces1
              -- eof
              return ss
       pattern = many (noneOf whitespacechars)
@@ -332,12 +335,21 @@ takeWidth w (c:cs) | cw <= w   = c:takeWidth (w-cw) cs
 -- (not counted), and line breaks (in a multi-line string, the longest
 -- line determines the width).
 strWidth :: String -> Int
-strWidth "" = 0
-strWidth s = maximum $ map (foldr (\a b -> charWidth a + b) 0) $ lines s'
-  where s' = stripAnsi s
+strWidth = maximum . (0:) . map (foldr (\a b -> charWidth a + b) 0) . lines . stripAnsi
 
+-- | Strip ANSI escape sequences from a string.
+--
+-- >>> stripAnsi "\ESC[31m-1\ESC[m"
+-- "-1"
 stripAnsi :: String -> String
-stripAnsi = regexReplace "\ESC\\[([0-9]+;)*([0-9]+)?[ABCDHJKfmsu]" ""
+stripAnsi s = case parseMaybe (many $ "" <$ try ansi <|> pure <$> anySingle) s of
+    Nothing -> error "Bad ansi escape"  -- PARTIAL: should not happen
+    Just xs -> concat xs
+  where
+    -- This parses lots of invalid ANSI escape codes, but that should be fine
+    ansi = string "\ESC[" *> digitSemicolons *> suffix <?> "ansi" :: Parsec CustomErr String Char
+    digitSemicolons = takeWhileP Nothing (\c -> isDigit c || c == ';')
+    suffix = oneOf ['A', 'B', 'C', 'D', 'H', 'J', 'K', 'f', 'm', 's', 'u']
 
 -- | Get the designated render width of a character: 0 for a combining
 -- character, 1 for a regular character, 2 for a wide character.
