@@ -27,10 +27,6 @@ module Hledger.Utils.Text
  -- isSingleQuoted,
  -- isDoubleQuoted,
  -- -- * single-line layout
-  textstrip,
-  textlstrip,
-  textrstrip,
-  textchomp,
  -- elideLeft,
   textElideRight,
  -- formatString,
@@ -54,12 +50,14 @@ module Hledger.Utils.Text
  -- fitStringMulti,
   textPadLeftWide,
   textPadRightWide,
+  -- -- * Reading
+  readDecimal,
   -- -- * tests
   tests_Text
   )
 where
 
--- import Data.Char
+import Data.Char (digitToInt)
 import Data.List
 #if !(MIN_VERSION_base(4,11,0))
 import Data.Monoid
@@ -78,22 +76,6 @@ import Hledger.Utils.Test
 -- lowercase = map toLower
 -- uppercase = map toUpper
 
--- | Remove leading and trailing whitespace.
-textstrip :: Text -> Text
-textstrip = textlstrip . textrstrip
-
--- | Remove leading whitespace.
-textlstrip :: Text -> Text
-textlstrip = T.dropWhile (`elem` (" \t" :: String)) :: Text -> Text -- XXX isSpace ?
-
--- | Remove trailing whitespace.
-textrstrip = T.reverse . textlstrip . T.reverse
-textrstrip :: Text -> Text
-
--- | Remove trailing newlines/carriage returns (and other whitespace).
-textchomp :: Text -> Text
-textchomp = T.stripEnd
-
 -- stripbrackets :: String -> String
 -- stripbrackets = dropWhile (`elem` "([") . reverse . dropWhile (`elem` "])") . reverse :: String -> String
 
@@ -420,6 +402,13 @@ textWidth s = maximum $ map (T.foldr (\a b -> charWidth a + b) 0) $ T.lines s
 --         | otherwise                        -> 1
 
 
+-- | Read a decimal number from a Text. Assumes the input consists only of digit
+-- characters.
+readDecimal :: Text -> Integer
+readDecimal = foldl' step 0 . T.unpack
+  where step a c = a * 10 + toInteger (digitToInt c)
+
+
 tests_Text = tests "Text" [
    test "quoteIfSpaced" $ do
      quoteIfSpaced "a'a" @?= "a'a"
